      program driver
      implicit none
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!
!  Sample driver for SPLIB, sparse matrix iterative library.  This sets
!  up the storage for A,x,b, and reads in a data file that drives the choice
!  of methods and various print/graphing options.
!
!  This version allocates all needed space dynamically.
!  This version also reads in the 
!  transpose of the matrix from a Harwell/Boeing format file.  The
!  reason for "transpose" is because H/B matrices are in compressed
!  sparse column format, while this driver treats them as in compressed
!  sparse row format.
!
!  This driver also reads in a sequence of solvers/preconditioners from
!  an input file called "methods" applying them until solver = 0 is
!  specified.  The xgraph files showing convergence history for each
!  method on a single run are written to the same file for comparisons.
!
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
    interface 
      subroutine readmt3(job,iounit,A,colind,rwptr,b,sol,g,nrhs, &
               guesol,nrow,ncol,nnz,title,key,rua,ierr,     &
               chkdiag)
      implicit none
      character title*72, key*8, rua*3, guesol*2
      integer nrow, ncol, nnz, neltvl, nrhs, nrwindx
      integer job, iounit, ierr
      double precision, pointer      :: A(:), b(:), sol(:), g(:)
      integer, pointer               :: rwptr(:), colind(:)
      logical chkdiag
    end subroutine
    end interface 

!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!        Documentation
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!
!  Solve a general sparse system Ax = b by means of calls to 
!  various iterative methods and preconditioners.  The driver will
!  loop over an input data file until it encounters a sentinal value,
!  and then quit.  The input matrix is read by Sparskit readmt.f,
!  modified to correctly handle the single right hand side case, and
!  modified to use pointers in a work array to define arrays read in.
!  A should be in the new Harwell/Boeing format, with right hand sides
!  getting a special format line of their own (the five line header
!  version of H/B format, not the four line header).
!
!  Because most of the preconditioners rely on the diagonal of the matrix
!  being present, a routine matchk then checks the data structure,
!  modifying it by inserting diagonal entries if necessary.  Those 
!  entries are set to zero; all that changes is the data structure,
!  not the matrix.
!
!------------------------------------------------------------------
!    
!     Randall Bramley
!     Department of Computer Science
!     Indiana University, Bloomington
!     bramley@cs.indiana.edu
!     Thu Jul 13 12:58:23 EST 1995
!
!------------------------------------------------------------------
!
!  Parameters:
!  ===========
!
!  mtot       : maximum number of integer words allowed for all problem
!               dependent arrays.  This includes the matrix, the vectors
!               defining the right hand side and solution, the preconditioners,
!               and the additional work vectors needed.
!
!  Variables (also see the descriptions in include file instr.inc):
!  ==============================================================
!
!  work    : Array of work storage, for preconditioners, solvers, and.
!            instrumentation routines.  The matrices and vectors used
!            are all indexed from this integer array, using pointers.
!            The following are all just integer pointers into work():
!
!  n       : Order of matrix A.
!  A       : (int pointer) double precision :: array for nonzero entries of 
!                          matrix A; of length nnz = rwptr(n+1)-1.
!  colind  : (int pointer) integer array of column indices for
!                          entries in A; of length nnz = rwptr(n+1)-1.
!  rwptr   : (int pointer) integer array of pointers to beginning position
!                          of rows in arrays A and colind.  Length n+1.
!
!  x       : (int pointer) double precision :: vector of unknowns to be solved for.
!  g       : (int pointer) double precision :: vector of initial estimate of x, if 
!                          provided by readmt3.
!  xs      : (int pointer) double precision :: vector of correct solution, if
!                          provided by readmt3.
!  b       : (int pointer) double precision :: right hand side vector of system to
!                          be solved: Ax = b
!            See lrhs and lsol for options on this vector.
!
!  next    : points to next available location in array work()
!
!  tol     : tolerance on norm used in stopping test.
!  ireltv  : variable to indicate whether stopping test
!            is relative to size of right hand side vector or not.
!            (0 for not, 1 to use relative residual).
!  rpcprm  : real preconditioner parameter -- 
!             - tolerance for numerical dropping with ILUT;
!             - omega parameter for SSOR(omega);
!             - relaxation constant for MILU.
!             - sparsity pattern indicator for ECIMGS
!  maxits  : Maximum number of iterations allowed.
!            On return from splib.f, actual number of iterations.
!  solmeth : Solution method to use.
!             solmeth = 1  ---> BiCG
!             solmeth = 2  ---> CG on normal equations AA'y = b, x = A'y.
!             solmeth = 3  ---> CG on normal equations A'A x = A'b
!             solmeth = 4  ---> CGS
!             solmeth = 5  ---> CGSTAB
!             solmeth = 6  ---> GMRES(k) 
!             solmeth = 7  ---> transpose-free QMR
!             solmeth = 8  ---> Bi-CG stabilized (Templates version of CG-Stab).
!             solmeth = 9  ---> Templates version of GMRES (still under work)
!             solmeth = 10 ---> Jacobi
!             solmeth = 11 ---> Gauss-Seidel
!             solmeth = 12 ---> SOR(omega)
!             solmeth = 13 ---> Orthomin(k)
!  krylov  : Size of Krylov subspace to use in GMRES or Orthomin; k above.
!  pcmeth  : Preconditioning method to use.
!             pcmeth = 0  ---> no preconditioner.
!             pcmeth = 1  ---> ILU(s)
!             pcmeth = 2  ---> MILU(s,rpcprm)
!             pcmeth = 3  ---> ILUT(s,rpcprm)
!             pcmeth = 4  ---> SSOR(rpcprm)
!             pcmeth = 5  ---> TRID(s), where s is the block size.
!             pcmeth = 6  ---> ILU(-1), the space saver version of ILU(0)
!             pcmeth = 7  ---> ECIMGS.
!  ipcprm  : integer preconditioning parameter, usually levels of fill.
!            For tridiag preconditioning, it means the block size.
!  n       : Order of matrix A, must be less equal to nmax.
!  nnz     : Number of nonzero entries in A, must be less equal 
!            to nzmax.
!  ind     : Integer used for loops and indexing.
!  nrhs    : Integer contains number of right hand sides returned by
!            readmt3.  Although multiple right hand sides may be read in,
!            currently the code will only solve the system for the first one.
!  job     : Integer to indicate what is to be read in readmt3.
!            Here only used for job = 2, i.e. read matrix only
!            and job = 3, i.e. read matrix and 1 right hand side
!  lrhs    : Logical variable indicating whether or not the rhs is supplied
!            in readmt3.  If it is not, an artificial one is created in setb.
!  lsol    : Logical variable indicating whether or not the true
!            solution vector is provided by readmt3.
!  lgus    : Logical variable indicating whether or not an initial guess for
!            solution vector x is provided by readmt3.
!
!  ierr    : Integer variable indicating error conditions.
!             ierr = -m --> Not enough workspace provided; m more double
!                           precision words needed in array work().
!             ierr = 0  --> Successful return; tolerance achieved.
!             ierr = 1  --> The initial guess satisfies the stopping test.
!                           This is often a clue that rhs not set properly.
!             ierr = 2  --> Tolerance reset 8 times in attempt to
!                           get true residual small enough.
!             ierr = 3  --> Convergence not achieved in maxits iterations.  
!             ierr = 4  --> Initial preconditioned residual is more than
!                           20 orders of magnitude larger than initial 
!                           unpreconditioned residual.  Not exactly an error
!                           condition, but when this happens the iterations never
!                           converge in practice.  Try another preconditioner.
!             4<ierr<9  --> breakdown condition of attempt to divide by
!                           near-zero number in algorithm.
!             Note that any of conditions 0, 1, 2 may mean success.
!
!  title   : Character array to hold title of Harwell/Boeing matrix.
!  key     : Character array for short title of Harwell/Boeing matrix.
!  guesol  : Character array for readmt.  Has interpretation
!            guesol(1:1) = 'G' or 'g' --> initial guess provided
!            guesol(2:2) = 'X' or 'x' --> correct solution provided.
!  res     : Final residual norm (2-norm)
!  err     : Final error norm (2-norm)
!  firstmat: Logical variable indicating if first time to execute splib,
!            so that matrix is only read in on first pass.  In common block
!            of instr.inc
!  fmtd    : Integer indicating if the matrix input file is formatted or not.
!              fmtd = 0 ---> formatted (ascii) input file
!              fmtd = 1 ---> unformatted (binary) input file
!
!  I/O Files
!  =========
!
!  methods : Input file for indicating method to be used, 
!            partitioning, etc.  This file has the repeated structure
!     
!     solmeth       
!     krylov
!     pcmeth
!     levelsfill
!     accel_parm
!     tol  ireltv
!     maxits
!     instlvl
!     iredo
!     
!     An example file "methods" to compare GMRES(15) and 
!     CGSTAB, both with MILU preconditioning might look like:
!     
!     5     !   Use CGSTAB iterative method.
!     0         !   Not used for CGSTAB ...
!     2     !   Use MILU preconditioning method
!     1     !   Number of levels of fill in preconditioner
!     1.0d0 !   Relaxation parameter for MILU.
!     1.0d-8 0  !   Stopping criterion of || A*x - b || < 1.0e-8
!     500   !   Allow up to 500 iterations
!     5         !   Provide full instrumentation
!     0         !   Scaling method (0 = none)
!     1         !   Input matrix is in formatted (ascii) file.
!     0         !   Recompute preconditioner 
!     
!     6     !   Use GMRES iterative method.
!     15    !   Use 15 Krylov vectors for GMRES 
!     2     !   Use MILU preconditioning method
!     1     !   Number of levels of fill in preconditioner
!     1.0d0 !   Relaxation parameter for MILU.
!     1.0d-8 0  !   Stopping criterion of || A*x - b || < 1.0e-8
!     500   !   Allow up to 500 iterations
!     5         !   Provide full instrumentation
!     0         !   Scaling method (0 = none)
!     1         !   Input matrix is in formatted (ascii) file.
!     1         !   Use same preconditioner as on previous run.
!     
!     0     !   Solmeth = 0 indicates stop run.
!
!  matrix  : Input file containing Harwell/Boeing matrix.  Actually
!            this collection solves a system with the transpose of
!            the H/B matrix provided, but a short routine to read 
!            in a H/B matrix and output its transpose is also 
!            available.  If fmtd = 0, then the file is assumed to be
!            in a binary file (Fortran unformatted).
!
!  results : Output file containing summary of results from run.
!
!  resi    : Output file with residual norm vs. iteration number.
!  rest    : Output file with residual norm vs. elapsed time.
!
!  reli    : Output file with relative residual norm vs. iteration number.
!  relt    : Output file with relative residual norm vs. elapsed time.
!
!  Mresi   : Output file with preconditioned residual norm vs. iteration
!            number.
!  Mrest   : Output file with preconditioned residual norm vs. elapsed
!            time.
!
!  Mreli   : Output file with preconditioned relative residual norm vs.
!            iteration number.
!  Mrelt   : Output file with preconditioned relative residual norm vs.
!            elapsed time.
!
!  oetli   : Output file with Oettli and Prager norm vs. iteration number.
!  oetlt   : Output file with Oettli and Prager norm vs. elapsed time.
!
!  errori  : Output file with error norm vs. iteration number
!            (if lsol = true).
!  errort  : Output file with error norm vs. time (if lsol = true).
!
!   ===================
!   Files opened/used:
!   ===================
!
!  The I/O unit numbers are provided in the vector iunit(20).  Note that units
!  4-6 are mandatory, and unit 20 can be changed to be the same as unit 6.
!  The other units are optional, depending on the error norm(s) you wish to
!  plot/use.
!  
!  For this driver the following assignments are made to iunit():
!
!  4 --> iunit(1): Input file for matrix in Harwell/Boeing format
!  5 --> iunit(2): Input file with control parameters
!  6 --> iunit(3): Output file with echoed input and summary of run(s).
!  
!  7 --> iunit(4): Output unit for || Ax - b ||  vs. iter
!  8 --> iunit(5): Output unit for || inv(M) (Ax-b) || 
!                  vs. iter
!  9 --> iunit(6): Output unit for || Ax - b || / || r_0 ||
!                  vs. iter
! 10 --> iunit(7): Output unit for || inv(M) (Ax-b) || / 
!                                  || inv(M) r_0 || vs. iter
! 11 --> iunit(8): Output unit for || Ax - b ||_inf /
!                                ( ||A||_inf ||x||_1 + ||b||_inf) vs. iter
! 12 --> iunit(9): Output unit for || x - xstar || vs. iter
! 13 --> iunit(10): Not used.
!  
! 14 --> iunit(11): Output unit for || Ax - b || vs. time
! 15 --> iunit(12): Output unit for || inv(M) (Ax-b) || vs. time
! 16 --> iunit(13): Output unit for || Ax - b || / || r_0 || 
!                   vs. time
! 17 --> iunit(14): Output unit for || inv(M) (Ax-b) || / 
!                   || inv(M) r_0 || vs. time
! 18 --> iunit(15): Output unit for || Ax - b ||_inf / 
!                   ( ||A||_inf ||x||_1 + ||b||_inf) vs. time
! 19 --> iunit(16): Output unit for || x - xstar || vs. time
!
!
!  =================
!  External Routines
!  =================
!
!  readmt3 : Reads Harwell/Boeing matrix into a, colind, rwptr format
!  bmux    : Sparse matrix multiply (Sparskit)
!  ddot    : Blas routine for dotproduct
!  dasum   : Blas routine for max norm of vector
!  dnrm2   : Blas routine for two-norm of vector
!
!  ==================
!  Workspace Required
!  ==================
!
!  The amount of workspace depends on the methods used, the levels
!  of fill-in (and way of defining fill-in levels), and the number
!  of Krylov subspace vectors used.  The following guide excludes 
!  the storage for x, r, and b which appear in the calling
!  sequence for all the routines.  Note that GMRES(k) depends
!  on the size k of the Krylov subspace.  Also note that the methods
!  with fill-in will require more storage for alu, jlu, ju.  The size
!  shown is for the preconditioners with no fill-in allowed, and does
!  not include the storage that depends on nnz.
!
!     Routine       Reals       Integers
!     -------       -----       --------
!
!     (m)ilus       n           2n
!     ilut          2n + 1      3n
!     ssor          1       n (but ju is recycled for this)
!     trid          4n      n
!     ilu0          n       n
!     ecimgs        ?       ?
!
!     Routine       Reals       
!     -------       -----       
!     bicg      6n              }
!     cgne      3n              }
!     cgnr      3n              }  Add 2n for all of these if
!     cgs       6n              }  the instrumentation routine
!     cgstab        6n              }  is to be used at all; the
!     gmres(k)      k*(n+k+5)+n + 1 }  extra is for work arrays
!     qmrtf         10n             }  used in residual computation.
!     bicgs     6n              }
!
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!
!   Declarations of variables:
!
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
      include 'include/symbols.inc'


!     --------------------------------
!     Double precision array pointers:
!     --------------------------------
      double precision, pointer :: A(:), x(:), b(:), xs(:), g(:), Dl(:), Dr(:)

!     -------------------------
!     Double precision scalars:
!     -------------------------
      double precision :: tol, rpcprm
      double precision :: err, work

!     ----------------------------
!     Function programs from BLAS:
!     ----------------------------
      double precision :: ddot,dasum,dnrm2
      external ddot,dasum,dnrm2

!     -----------------------
!     Integer array pointers:
!     -----------------------
      integer, pointer ::  rwptr(:), colind(:)
      integer, pointer ::  diag(:)
      integer next

!     ---------
!     Integers:
!     ---------
      integer ierr, maxits, iparms(2), n, nnz
      integer i, k, mneed
      integer job, nrhs, ipcprm
      integer pcmeth, solmeth, krylov, ndiag
      integer digprec
      integer fmtd
      integer scaling
      integer alloc_stat

!     ------------------
!     Logical variables:
!     ------------------
      logical chkdiag
      logical, save :: firstmat
      data firstmat /.true./
      logical lsol, lrhs, lgus, reltv, redo
      integer ireltv, iredo

!     --------------------
!     Character variables:
!     --------------------
      character rua*3, guesol*2
      character title*72, key*8
      character solver*5, precond*4
      character from*8

!     --------------------------
!     Instrumentation variables:
!     --------------------------
      integer iunit(20), instlvl
      common /truesol/ xstar
      double precision :: xstar(1)

!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!      Beginning of Executable Statements:
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

!     ----------------------------------------------
!     Input file for matrix in Harwell/Boeing format
!     ----------------------------------------------
      iunit(3) = 4

!     ----------------------------------
!     Input file with control parameters
!     ----------------------------------
      iunit(2) = 5

!     ----------------------------------------------------
!     Output file with echoed input and summary of run(s).
!     ----------------------------------------------------

      iunit(1) = 60

!     ------------------------------------------------
!     Other unit number somewhat arbitrarily assigned:
!     ------------------------------------------------
      do i = 4,20
         iunit(i) = i+3
      end do 

!     -----------------------------------------
!     Open required files for input and results
!     -----------------------------------------
      open(iunit(2),file='methods')
      open(iunit(1),file='results.html')

!*****************************************************************************
!****** Big "go to" loop over problem solution/preconditioner specifications:*
!*****************************************************************************

  20  continue
      ierr = 0

!     -----------------------------------------------------------------
!     Read in parameters that define stopping tolerances, methods to be
!     used.  Unpack information into pcmeth, ipcprm, etc.  Open 
!     additional I/O files if specified by instlvl.
!     -----------------------------------------------------------------
      call unpack (ireltv,   rpcprm    ,pcmeth  ,           &
                   ipcprm   ,solmeth ,krylov  ,             &
                   tol     ,maxits  , iredo,                &
                   iunit   ,                                &
                   instlvl , scaling, fmtd,                 &
                   title, key, solver  , precond)

!     ------------------------------------------------------
!     Try to read right hand side and solution vector.
!     If not supplied, create the right hand side vector b.
!     ------------------------------------------------------
      job = 3

!     -----------------------------------------------------------------------
!     Read matrix and vectors, and set up pointers.  Note that readmt3 will
!     insert extra positions after the colind array, to provide for 
!     the possible expansion of data structure by inserting missing diagonal
!     entries.
!     -----------------------------------------------------------------------
      if (firstmat) then

!       ------------------------------------------------------------
!       Force matrix data structure to have diagonal entries present
!       ------------------------------------------------------------
        chkdiag = .true.
        if (fmtd  ==  0) then
           ! --------------------------
           ! Use unformatted input file
           ! --------------------------
           job = 0
           open(iunit(3),file='matrix',form='unformatted')
           call readunf2(n,n,nnz,A,colind,rwptr,iunit(3), job,      &
             ierr,nrhs,b,title,key,rua,digprec)
           allocate (A(nnz), x(n), b(n), g(n), xs(n), stat = alloc_stat)
           if (alloc_stat /= 0) stop "Could not allocate A"
           allocate (colind(nnz), rwptr(n+1), stat = alloc_stat)
           if (alloc_stat /= 0) stop "Could not allocate rwptr, colind"
           job = 3
           rewind(iunit(3))
           call readunf2(n,n,nnz,a,colind,rwptr,iunit(3), job,      &
             ierr,nrhs,b,title,key,rua,digprec)
        else
           ! ----------------------------------------------------------
           ! Use formatted input file.
           ! First, read header data to get sizes for memory allocation
           ! ----------------------------------------------------------
           job = 0
           open(iunit(3),file='matrix',form='formatted')
           call readmt3(job,iunit(3),A,colind,rwptr,b,xs,g,nrhs, &
              guesol,n,n,nnz,title,key,rua,ierr,chkdiag)
           allocate (A(nnz), x(n), b(n), g(n), xs(n), stat = alloc_stat)
           if (alloc_stat /= 0) stop "Could not allocate A"
           allocate (colind(nnz), rwptr(n+1), stat = alloc_stat)
           if (alloc_stat /= 0) stop "Could not allocate rwptr, colind"
           ! -----------------------
           ! Now read in matrix data 
           ! -----------------------
           job = 3
           rewind(iunit(3))
           call readmt3(job,iunit(3),A,colind,rwptr,b,xs,g,nrhs, &
               guesol,n,n,nnz,title,key,rua,ierr,chkdiag)
           deallocate (g)
        end if

        if (ierr .ne. 0) then
             write(iunit(1),5000) ierr
             write(iunit(1),5010) mneed
             stop
        end if 
        lrhs = (job .ge. 3)
        lsol = (guesol(2:2)  ==  'X' .or. guesol(2:2)  ==  'x')
        lgus = (guesol(1:1)  ==  'G' .or. guesol(1:1)  ==  'g')

!       --------------------------------------------------------------
!       Call matrix checker to insert missing diagonal entries, and to
!       sort rows into increasing column order:
!       --------------------------------------------------------------
        allocate(diag(n), stat=alloc_stat)
        if (alloc_stat /= 0) stop "Could not allocate diag in driver"
        call matchk(A,colind,rwptr,n,diag, ndiag)

!       ------------------------------------------------------
!       If job < 3, the right hand side vector must be set up.
!       ------------------------------------------------------
        if (job < 3) call setb(n,b)

        ! Special insert, to remove constant vector from rhs:
        ! work = dasum(n,b,1)  ! This is b'e, e = (1,1,...,1)'
        ! do k = 1, n
        !     b(k) = b(k) - work
        ! end do


      end if ! firstmat


!     ----------------------------------------------------
!     Echo information about matrix and sol/precon methods
!     ----------------------------------------------------
!      write(iunit(1),6070) title(1:72)
!      write(iunit(1),6080) key

!     ---------------------------------------------------------
!     Copy over true solution vector to xstar, if it was
!     provided in readmt3.  If it was not, then we cannot
!     check the error norm, so set that in the errchk vector
!     in case the user tried to specify it.  Note: since
!     having a true solution vector is rare, in this driver
!     and in splib, it is passed using the common block truesol.
!     If you wish to see the true error norms, you must change
!     the declaration of the vector xstar in common truesol in
!     both driver.f and instr.f, recompile, and then uncomment 
!     out the next five lines.
!     ---------------------------------------------------------
!      if (lsol) then
!          call dcopy(n,xs,1,xstar(1),1)
!      else
!          instlvl  = min(instlvl, 5)
!      end if 
      instlvl  = min(instlvl, 5)

!     --------------------------------------------------------
!     Set up initial x appropriately.  If a "guess" vector was
!     provided in readmt3, use it.  Otherwise, set x randomly.
!     --------------------------------------------------------
      if (lgus) then
         call dcopy(n,g,1,x,1)
      else
         call random_number(x(1:n))
      end if

!     -----------------------------------
!     Pack up the iparms vector for splib
!     -----------------------------------
      iparms(1) = ipcprm
      iparms(2) = krylov
      if (.not. associated(Dl)) allocate(Dl(n), Dr(n), stat = alloc_stat)
      if (alloc_stat /= 0) stop "Could not allocate Dl or Dr"

!     ----------------------------
!     Call splib to apply methods.
!     ----------------------------
      call splib(A, colind, rwptr, n,             &
                 x, b,                            &
                 tol, ireltv, maxits, pcmeth, solmeth,        &
                 scaling, Dl, Dr,                 &
                 iparms, rpcprm, iredo, ierr, iunit, instlvl)

!      open(77,file='x')
!      call prtval(77,'dbl',n,x,x)

!     ---------------------------------------------
!     Compute error norm if true solution available
!     ---------------------------------------------
      if (lsol) then
          allocate(g(n), stat=alloc_stat)
          if (alloc_stat /= 0) stop "Could not allocate g in driver"
          call dcopy(n, x, 1, g, 1)
          call daxpy(n, -1.0d0, xs,1, g,1)
          err = dnrm2(n,g,1)
          write(iunit(1),2010) err
          deallocate(g)
      end if

!     -----------------------------------------------
!     Pick up next solver/preconditioner pair to try.
!     -----------------------------------------------
      firstmat = .false.

      go to 20

!========================================================================
!           Format Statements
!========================================================================

 6020  format('Order of matrix = ',i6,/,'Number of nonzeros ',i8)
 6040  format('Tolerance on residual = ',1pe10.4)
 6050  format('Iterative method used: ',a8,' with ',a8, ' preconditioning')
 6060  format('Levels of fill allowed in precond: ',i3)
 6070  format(/,40('--'),/,'Problem title: ',a71)
 6080  format('Problem key: ',a8)
 6090  format('Acceleration parameter: ',f8.2)

 2010  format('Final error 2-norm: ',1pe12.5)

 5000  format(' *** Error in readmt ; ierr = ',i4)
 5010  format(' *** Error in readmt ; mneed = ',i7)

    end
!----------------------------------------------------------------------
