!========================================================================
      subroutine lusl (n, y, x, lu, jlu, uptr, pcmeth, A, colind, rwptr) 

!-----------------------------------------------------------------------
!                     *** LUSL: Solve Mx = y ***            
!                                   
!   Preconditioner application for SPLIB.  This routine computes the    
!   action of the preconditioner on a vector; usually this is a lower   
!   and upper triangular solve, explaining the name.  In particular,    
!   (lu, jlu) is usually the MSR data structure for L and U combined    
!   with the integer vector uptr giving the pointers to the start   
!   position of the upper triangular part of each row.  However, for    
!   SSOR preconditioning lu is the scalar relaxation parameter omega    
!   and uptr is a vector of pointers to the diagonal entries in the 
!   data structure for A (not LU). In this case the names lu and jlur   
!   are somewhat deceptive, but have been kept since ILU, MILU, ILUT,   
!   and ECIMGS all use the MSR data structure for LU.           
!                                   
!   Although x and y can be different, the methods implemented so far   
!   allow overwriting, so there is an inefficiency in copying y over    
!   to x unnecessarily.                         
!                                   
!-----------------------------------------------------------------------
!
!  Arguments:
!  ==========
!
!  x       : double precision :: vector of length n (output).
!  y       : double precision :: vector of length n (input).
!
!  n       : Order of matrix A, and system Mx = y where M is the
!            preconditioning matrix.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!  uptr, 
!  lu,jlu  : Data structure for the preconditioner M, with different meanings
!            for different preconditioners.
!
!            ILU(s), MILU(s), ILUt, or ECIMGS --->
!                 (lu,jlu) gives the combined L/U factors for the preconditioner
!                 in a modified sparse row (MSR) format.  L is unit lower 
!                 triangular and U is upper triangular.  The diagonal of U is
!                 stored in lu(1:n) inverted, to replace divisions in the algorithm
!                 with multiplications.  Each i-th row of the combined L/U matrix in
!                 the (lu,jlu) data structure contains the i-th row of L (excluding
!                 the diagonal entry = 1), followed by the i-th row of u.
!                 uptr is a integer vector of length n containing pointers to the
!                 start position of each row of U in (lu, jlu).
!
!            SSOR(omega)   --->
!                 lu is the scalar acceleration parameter omega.
!                 jlu is not used, and uptr is a vector of pointers to the
!                 diagonal entries in the CSR data structure for A.
!
!            TRID(s)   --->
!                 This implements block diagonal preconditioning, where
!                 each diagonal block is from the corresponding tridiagonal part
!                 of the matrix A.  The size of the blocks is given implicitly in
!                 the integer iskip, stored in jlu(n).  Factors are set up for 
!                 contiguous blocks of iskip rows and so by using iskip = n the 
!                 factor can be for the entire matrix's tridiagonal part.
!                 Using iskip = 1 is an inefficient way of performing diagonal
!                 preconditioning, while 1 < iskip < n is for a block tridiagonal
!                 preconditioner.  Gaussian elimination with pivoting is used for 
!                 factoring each block, so it is necessary to store the pivot
!                 sequence in jlu(1:n-1).  lu has the four vectors giving the nonzeros
!                 of the tridiagonal preconditioner.  Vector uptr is not used.
!
!            ILU0   --->
!                 This implements the version of ILU(0) that does not change any
!                 off-diagonal entries of A, and so only needs to store the n-vector
!                 of diagonal pivots.  lu(1:n) stores those pivots, and jlu(1:n) 
!                 contains pointers to the diagonal entry in each row of A.
!
!  pcmeth  : preconditioning method to use.
!       pcmeth = 0  ---> no preconditioner, i.e. m = identity matrix.
!       pcmeth = 1  ---> ILU(c).
!       pcmeth = 2  ---> MILU(c).
!       pcmeth = 3  ---> ILUT(c).
!       pcmeth = 4  ---> SSOR(omega).
!       pcmeth = 5  ---> TRID.
!       pcmeth = 6  ---> ILU0.
!       pcmeth = 7  ---> ECIMGS.
!
!
!  Routines
!  ========
!
!  lusolt  : performs a forward and a backward solve for matrices L and U
!            stored in combined MSR format, with L unit upper triangular.
!            the diagonal entries of U are assumed to be inverted already.
!  lussor  : Performs forward and backward solve for SSOR preconditioner.
!  lusol0  : Performs forward and backward solve for ILU0 preconditioner.
!  tridsl  : Applies the tridiagonal block preconditioner.
!  dcopy   : BLAS1 routine to copy a vector.
!-------------------------------------------------------------------------
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University, Bloomington
!     bramley@cs.indiana.edu
!     Wed Jul 12 10:03:18 EST 1995
!
!-------------------------------------------------------------------------

      implicit none

!     Declarations of arguments
!     =========================
      double precision :: y(*),x(*),lu(*)
      double precision :: a(*)
      integer n
      integer jlu(*),uptr(*),pcmeth
      integer colind(*),rwptr(n+1)

!     Local variables
!     ===============
      integer k, length, ierr

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!========================================================================
!      Beginning of Executable Statements
!========================================================================

      tstart = mytime()

!     ----------------
!     Copy y over to x
!     ----------------
      call dcopy(n, y, 1, x, 1)

!     ------------------------------------------
!     Unknown pcmeth (should not be encountered)
!     ------------------------------------------
      if (pcmeth <  0 .or. pcmeth  >   7) then
         write(*,*) 'Unknown method in lusl'
         write(*,*) 'pcmeth = ', pcmeth
         stop
      end if

!     ------------------
!     No preconditioning
!     ------------------
      if (pcmeth == 0) then
         return
      endif

!     --------------------------
!     ILU, MILU, ILUT, or ECIMGS
!     --------------------------
      if ( ((pcmeth >= 1) .and. (pcmeth <= 3)) .or. pcmeth  ==  7)  &
             call lusolt (n, x, lu, jlu, uptr) 

!     --------------------
!     SSOR preconditioning
!     --------------------
      if (pcmeth == 4) then
        call lussor(n, x, lu, jlu, A,colind,rwptr ) 
      end if

!     --------------------
!     ILU0 preconditioning
!     --------------------
      if (pcmeth == 6) call lusol0 (n, x, lu, jlu, A,colind,rwptr ) 

!     --------------------
!     TRID preconditioning
!     --------------------
      if (pcmeth == 5) then
         do 100 k = 1, n, jlu(n)
            length = min(jlu(n),n-k+1)
            call tridsl('N',length,lu(1+k),             &
                                   lu(k+n),             &
                                   lu(k+2*n),           &
                                   lu(k+3*n),           &
                                   jlu(k),x(k),ierr)
 100     continue
      end if

      times(5) = times(5) + mytime() - tstart

      return

      end subroutine lusl

!=======================================================================
    subroutine lusolt (n, x, lu, jlu, uptr) 
!-----------------------------------------------------------------------
!
! Perform a forward then backward solve for a MSR matrix containing
! a unit lower triangular and an upper triangular matrix with inverted
! diagonal, both stored in a single MSR data structure.  The vector of
! pointers uptr() contains the row pointers to the beginning of U in
! each row. 
!
!-----------------------------------------------------------------------
        implicit none
        integer jlu(*),uptr(*),n,i,k 
        double precision :: x(n), lu(*)

!       ------------
!       Solve Lx = x
!       ------------
        do i = 1, n 
           do k = jlu(i), uptr(i)-1 
              x(i) = x(i) - lu(k)* x(jlu(k)) 
           end do
        end do

!       ------------
!       Solve Ux = x
!       ------------
    do i = n, 1, -1 
       do k = uptr(i), jlu(i+1)-1 
              x(i) = x(i) - lu(k)*x(jlu(k))
           end do
           x(i) = lu(i)*x(i)
        end do

    return
    end subroutine lusolt

!----------------------------------------------------------------------- 
      subroutine lusol0(n, x, pivots, diag, A,colind,rwptr) 
!----------------------------------------------------------------------- 
!
!  Solve the system LUx = x, where L and U were created using
!  ILU(-1) preconditioner, the space saver version of ILU(0).
!  The implicit splitting is
!          LU = (D + LA) (I + (D)^-1 UA)
!
!----------------------------------------------------------------------- 
      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      double precision :: x(n), pivots(n)
      double precision :: A(*), sum
      integer i, j

      do i = 1, n
         sum = 0.0d0
         do j = rwptr(i), diag(i)-1
            sum = sum + A(j)*x(colind(j))
         end do
         x(i) = pivots(i)*(x(i) - sum)
      end do

      do i = n, 1, -1
         sum =  0.0d0
         do j = diag(i)+1, rwptr(i+1)-1
            sum = sum + A(j)*x(colind(j))
         end do
         x(i) = x(i) - pivots(i)*sum
      end do

      return
      end subroutine lusol0

!-----------------------------------------------------------------------
      subroutine lussor(n, x, omega, diag, A,colind,rwptr ) 
!-----------------------------------------------------------------------
!
!  Solve the two triangular systems given for the SSOR preconditioner
!  with acceleration parameter omega.
!
!  System solved is 
!      Mx = y, with M = (D/omega + L) * ( I + omega * D^-1 * U).
!
!  diag : vector of pointers to the diagonal entry in each row k of A.
!  Warning: the rows of A must be sorted so that within each row,
!           the subdiagonal entries come first, then the diagonal entry,
!           and then the superdiagonal entries.  Furthermore, the diagonal
!           of A must be nonzero.
!
!-----------------------------------------------------------------------

      implicit none
      integer n, colind(*), rwptr(n+1), diag(n)
      double precision :: x(n)
      double precision :: A(*), omega, sum
      integer j, k

      do k = 1, n
         sum = 0.0d0
         do j = rwptr(k), diag(k)-1
            sum = sum + A(j)*x(colind(j))
         end do
         x(k) = (x(k)-sum) * omega / a(diag(k))
      end do

      do k = n, 1, -1
         sum = 0.0d0
         do j = diag(k)+1, rwptr(k+1)-1
            sum = sum + A(j)*x(colind(j))
         end do
         x(k) = x(k) - sum * omega / a(diag(k))
      end do

      return
      end subroutine lussor
