!======================================================================== 
      subroutine  numfac(n, colind,rwptr,  jlu, uptr, a, lu,        &
                    relax, ierr, colptrs, milu)

!======================================================================== 
!
! Numerical factorization with given sparsity pattern identified in
! the MSR data structure  jlu.
!
! Because of the MSR storage format, the method used is a deferred update
! version.  At the k-th step, the updates from the previous rows m,
! m < k and m corresponding to a nonzero in row k of L/U, are applied
! to row k.  A relaxation parameter is used.  When it is 0.0, ILU
! factorization is performed.  Otherwise the parameter is multiplied
! times the discarded fillin and added to the diagonal entry.  So when 
! the parameter is equal to 1.0, classical MILU factorization is performed.
!
! This routine implicitly assumes that the sparsity pattern specified
! by the MSR data structure contains that of A.  Note the comments in
! the code if you want to remove this restriction.
!
! Matrix A is input in CSR format, and the output MSR array contains
! both L and U, with L assumed to be unit lower triangular.  The diagonal
! entries of U are stored in inverted form to avoid divisions when
! applying the preconditioner.
!
! This version differs from earlier SPLIB versions in performing the
! computations in-place, without using a software accumulator register
! row(:).  The reasoning is that for very large systems, the memory
! access patterns when using such a register will cause many cache
! misses.  The technique (suggested by Xiaoge Wang) is to use an integer
! vector colptrs which simultaneously identifies which columns correspond
! to allowed fill by being nonzero, and stores the index within the data
! structure for L/U of that allowed fill.
!
! Finally, if a small pivot element is found in the factorization, it
! is replaced with a small quantity based on machine epsilon, the sign
! of the pivot element, and the norm of the current row.
!
!======================================================================== 
!   Matlab code for computing this factorization is:
!
!      U = zeros(n,n);
!      L = eye(n,n);
!   
!      U(1,1:n) = A(1,1:n);
!   
!      for k = 2:n
!         L(k,1:k-1) = (U(1:k-1,1:k-1)'\(A(k,1:k-1)'))';
!         U(k,k:n) = A(k,k:n) - L(k,1:k-1)*U(1:k-1,k:n);
!      end ;
!   
!   
!======================================================================== 
!
!  The process can be viewed best in the dense case, shown here
!  for k = 4 and n = 6.  Rows 1 through 3 have been computed and
!  are complete at the beginning of this stage (they need no
!  further updates).  The  corresponding dense factorization is
!  usually viewed as consisting of a triangular solve to find the
!  entries L(k,1:k-1), followed by a vector*matrix update to get
!  the entries U(k,k:n).  When a saxpy form is used for both the
!  triangular solve and the vector*matrix update, both phases have
!  the same form and can be computed with the same loop.
!  
!  Stage 0: Copy row 1 of A over to row 1 of LU.
!
!  Stage 1: Compute LU(4,1), as LU(4,1)/U(1,1).
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           | slv |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 2: Update LU(4,2:n) using multiplier in LU(4,1) and row 1 of LU:
!  
!           -------------------------------------
!           |     | U12 | U13 | U14 | U15 | U16 |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           | M   | upd | upd | upd | upd | upd |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 3: Compute LU(4,2), as LU(4,2)/U(2,2).
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     | slv |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 4: Update LU(4,3:n) using multiplier in LU(4,2) and row 2 of LU:
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     | U23 | U24 | U25 | U26 |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     | Ml  | upd | upd | upd | upd |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 5: Compute LU(4,3), as LU(4,3)/U(3,3).
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     | slv |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 6: Update LU(4,4:n) using multiplier in LU(4,3) and row 3 of LU:
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     | U34 | U35 | U36 |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     | M   | upd | upd | upd |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stage 7: Invert diagonal entry U(4,4), which is now complete:
!  
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     | Inv |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!           |     |     |     |     |     |     |
!           |     |     |     |     |     |     |
!           -------------------------------------
!  
!  Stages 8 onwards: Go on to the next row, repeating above process.
!
!  Note:  When applying updates from previous rows to the
!  current one, we only need apply those for which there is a
!  corresponding nonzero in the current row of LU; otherwise the
!  multiplier would be zero.
!
!     Randall Bramley and Xiaoge Wang
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Sun Jun 26 09:51:56 EST 1994
!  
!======================================================================== 
!     a, colind, rwptr    :  Matrix A in CSR format
!     lu, jlu, uptr: Data structure for L/U matrix in MSR format.
!                    lu contains the nonzero values of L/U
!                       lu(1:n) diagonal entries listed in order
!                       lu(n+1) is unused,
!                       lu(n+2:nzlu+1) are the off-diagonal entries in
!                               column increasing order.
!                    jlu contains pointers and column indices.
!                       jlu(1:n) pointers to the beginning of each row
!                                in arrays lu and jlu
!                       jlu(n+1) is a pointer to one beyond the last
!                                entry in lu and jlu
!                       jlu(n+2:nnz+1) are the column indices
!                    uptr(1:n) contains pointers the first entry of U in
!                                each row.
!     n             : order of the matrices.
!     relax         : relaxation parameter for MILU methods.
!     mult          : temporary scalar for holding multiplier L(k,j).
!     ierr          : return error code.
!                        ierr = 0  -> all's OK.
!                        ierr < 0  -> row -ierr had a small pivot
!     k             : loop index for current row number
!     j             : loop index for current row number
!     indj          : index for traversing row k; index of j-th entry in
!                        the lu and jlu arrays.
!     indja         : index for traversing row k of A.
!     inds          : index for traversing updating row of L/U when 
!                        processing row k by applying sparse saxpy ops from
!                        previous (updating) rows to row k (updated row).
!     jluj          : index j of current column number in row k
!     jlus          : index s of current column number in row j
!     ijaj          : column index j for entries of A.
!
!  colptrs(1:n)     : used as an indicator array, to perform updates only
!                     on columns that corr to allowed nonzeros.   If
!                     column j is an allowed nonzero entry of the
!                     current row, then colptrs(j) is the
!                     index in arrays lu(:) and jlu(:) of
!                     the corresponding entry (k,j) of L/U
!  milu             : logical indicating whether to relax diagonal entries
!                     or not
!  rwnrm            : row norm of current row; used for determining small
!             diagonal element replacement.
!  nrw              : number of nonzeros in current row of A.
!
!======================================================================== 

      implicit none

!     ------------------------
!     Declaration of arguments
!     ------------------------
      integer  n, colind(*), rwptr(*), jlu(*), uptr(*)
      double precision ::   a(*), lu(*), relax
      integer  colptrs(n), ierr, nrw

!     ---------------
!     Local variables
!     ---------------
      integer  k, indj, inds, indja
      integer  jluj, jlus, ijaj
      logical  milu
      double precision ::  SMALL
      double precision ::  rwnrm, mult

!========================================================================
!      Beginning of Executable Statements
!========================================================================

      SMALL = sqrt(epsilon(SMALL))

!     ------------------------------------------------------------
!     colptrs is used to hold the indices of entries in LU of 
!     row k.  It is initialized to zero here, and then reset after 
!     each row's work.
!     ------------------------------------------------------------
      do k =  1, n
         colptrs( k ) = 0
      end do

!     ------------------
!     Proceed row by row
!     ------------------
      do 200 k = 1, n

!        --------------------------------------------------------------
!        Set up colptrs with indices in lu of allowed nonzeros of row k
!        --------------------------------------------------------------
         do indj = jlu(k), jlu(k+1)-1
            colptrs(jlu(indj)) = indj
            lu(indj) = 0.0d0
         end do

!        ---------------------------------------------------------
!        Set the diagonal entry (not needed for CSR format of ILU)
!        ---------------------------------------------------------
         colptrs(k) =  k

!        ----------------------------------------------------------------
!        Copy row k of A over to LU.  Note that the "if" test in the loop
!        can be removed if it is known that the sparsity pattern of A
!        is contained in that of LU, which is the case for (M)ILU(s).
!        ----------------------------------------------------------------
         rwnrm = 0.0d0
         nrw = 0
         do indja = rwptr(k), rwptr(k+1)-1
            ijaj = colind(indja)
!            if (colptrs(ijaj) .ne. 0) then
                nrw = nrw + 1
                rwnrm =  rwnrm + abs(a(indja))
                lu(colptrs(ijaj)) = a(indja)
!            end if
         end do

!        -------------------------------------------------------------------
!         The first segment of the next loop on indj effectively solves
!         the transposed upper triangular system
!                 U(1:k-1, 1:k-1)'L(k,1:k-1)' = A(k,1:k-1)'
!         via sparse saxpy operations, throwing away disallowed fill.
!         When the loop index indj reaches the k-th column (i.e., the
!         diagonal entry), then the innermost sparse saxpy operation 
!         effectively is applying the previous updates to the corresponding 
!         part of U via sparse vector*matrix, discarding disallowed fill-in
!         entries.  That operation is 
!            U(k,k:n) = A(k,k:n) - U(1:k-1,k:n)*L(k,1:k-1)
!        -------------------------------------------------------------------

         do 100 indj = jlu(k), uptr(k)-1


!           -----------------------------------------------------------
!           jluj is the col number of current entry in row k of L,
!           and index of diagonal entry of U in same column.  For LU in
!           CSR format, that diag entry will require fancier indexing.
!           -----------------------------------------------------------
            jluj = jlu(indj)

!           -----------------------------------------------------------
!           Solve for next unknown in row k of L: L_kj = L_kj/U_jj ...
!           -----------------------------------------------------------
            lu(indj) = lu(indj)*lu(jluj)
            mult = lu(indj)

!           -------------------------------------------------------------
!           ... and use it as a multiplier to update the entries s in row
!           k of L, s = j+1, ... , k-1, and the entries s in row k of U,
!           s = k, ..., n.
!           -------------------------------------------------------------
            if (milu) then
              do inds = uptr(jluj), jlu(jluj+1)-1
                 jlus = jlu(inds)
                 if (colptrs(jlus) .ne. 0) then
                    lu(colptrs(jlus)) = lu(colptrs(jlus)) - mult*lu(inds)
                 else
                    lu(k) = lu(k) - relax*mult*lu(inds)
                 end if
              end do
            else
              do inds = uptr(jluj), jlu(jluj+1)-1
                 jlus = jlu(inds)
                 if (colptrs(jlus) .ne. 0) then
                    lu(colptrs(jlus)) = lu(colptrs(jlus)) - mult*lu(inds)
                 end if
              end do
            end if

 100     continue


!        ----------------------------------------------------------
!        Finished with row k of LU; reset colptrs indices to zero 
!        for next row, and invert diagonal entry of U for this row.
!        ----------------------------------------------------------
         do indj = jlu(k), jlu(k+1)-1
            colptrs(jlu(indj)) = 0
         end do

         colptrs(k) =  0

         if (abs(lu(k)) .le. SMALL*rwnrm/nrw) then
            lu(k) = sign(SMALL*rwnrm/nrw, lu(k))
!            ierr = -k
!            return
         end if
         lu(k) = 1.0d0/lu(k)

 200  continue

      ierr  = 0
      return

      end subroutine numfac

