!======================================================================== 
        subroutine ssor (n,A,colind,rwptr,omega,lu,uptr,storg,ierr)

!*********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    *
!                     Indiana University                             *
!*********************************************************************

!======================================================================== 
!                      *** SSOR preconditioner ***          
!                                   
!  Set up implicitly the SSOR preconditioning matrix.  Let      
!  A = L + D + U be the usual splitting, with D diagonal, L strictly    
!  lower triangular, and U strictly upper triangular.  Because an   
!  additional matrix for L and U need not be created, this routine  
!  simply sets up pointers to the diagonal entries of A for use     
!  in the triangular solves needed.  It also orders the row entries     
!  of A.                                
!                                   
!  The preconditioner is given by                   
!      M = (I+omega*L*D^-1) * (D/omega+U)               
!  where omega in (0,2) is the SSOR acceleration parameter.     
!  Note that the usual way of writing this is as            
!     M = (1/(2-omega)) * (D/omega+L)*(omega*D^-1) * (D/omega+U)    
!  and so I have left off the scaling factor 1/(2-omega) and have   
!  combined the last two factors.  The second form shows that when  
!  A is symmetric, so is the SSOR preconditioner.           
!                                   
!  The calling sequence of this routine has lotsa garbage that is   
!  not needed, to keep it consistent with the explicit setup of     
!  SSOR in an earlier version.                                      
!                                   
!  Note: the diagonal of A must be larger than the parameter        
!        SMALL, set in this routine as a parameter.         
!                                   
!     Randall Bramley                           
!     Department of Computer Science                    
!     Indiana University                        
!     email: bramley@cs.indiana.edu                 
!     Wed Jun 22 14:57:13 EST 1994                  
!                                   
!======================================================================== 
!                                   
! PARAMETERS
!-----------
!
! on entry:
!========== 
! n       = integer. The dimension of the matrix A.
!
! a,colind,rwptr = matrix stored in Compressed Sparse Row format.
!
! work    = dummy work array 
!
! omega   = double precision :: relaxation parameter omega for ssor.
!           Theory requires  0 < omega < 2.
!
! On return:
!=========== 
!
! lu   = parameter omega
! uptr = integer array of length n containing pointers to        
!           the diagonal entry of each row of A.
!                                                                       
! storg   = integer. Amount of double precision :: storage needed for precond (always 1)
!
!  ierr is an error flag:
!        ierr  = -i --> near zero pivot in step i
!        ierr  = 0  --> all's OK
!        ierr  = 1  --> not enough storage; check mneed.
!        ierr  = 2  --> illegal parameter
!
!======================================================================== 

        implicit none

!       Declaration of arguments
!       ========================
        integer rwptr(*), colind(*), uptr(*), n, ierr
        double precision :: A(*), lu, omega
        integer storg

!       Local variables
!       ===============
        integer i,k,is,ie
        double precision :: SMALL


!=======================================================================
!     Start of executable statements
!=======================================================================

!       -------------------------------
!       Set tolerance for safe division
!       -------------------------------
        SMALL = sqrt(epsilon(SMALL))

        storg = 1
        ierr = 0

!       ------------------------------
!       Check for omega in right range
!       ------------------------------
        if (omega  <=  0.0 .or. omega  > 2.0) then
           ierr = 2
           return
        end if

        lu = omega

!       ----------------------------
!       Proceed row by row through A
!       ----------------------------

        do 60 k = 1, n

!          -----------------------------------------
!          Start indices of rows k and k+1 in matrix
!          -----------------------------------------
           is = rwptr(k)
           ie = rwptr(k+1)

!          ------------------------------------------------
!          Sort k-th row into increasing column index order.
!          ------------------------------------------------
           call rowinc( ie - is, colind(is), A(is) )

!          ------------------------------
!          Find pointer to diagonal entry
!          ------------------------------
           uptr(k) = 0

           do 20 i = is, ie-1
              if (colind(i)  ==  k) then
                 uptr(k) = i
                 go to 40
              end if
 20        continue

!          --------------------------------------------
!          Error condition if diagonal entry is missing
!          --------------------------------------------
           ierr = -k
           return

 40        continue

!          ----------------------------------------------
!          Error condition if diagonal entry is near-zero
!          ----------------------------------------------
           if (abs(A(i))  <=  SMALL) then
              ierr = -k
              return
           end if

 60     continue

    return
    end         
