      subroutine symbgs(n, A, c_ind , r_point, omega,       &
                         ilu   ,ilup  ,  iwork , space  ,   &
                         ierr  , mneed )

      implicit none
      include '../include/intdbl.inc'
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
!                                                                       
!    Method:                                                            
!                                                                       
!  The code consists of two parts:                  
!      (1) Pattern selection using parameter omega. When Omega = 0      
!  the pattern of target ECIMGS factors is the pattern of matrix A.     
!  When 0 < omega < 1.0, the pattern of target factors is sparser       
!  then the give matrix A. It use omega as the tolerance to drop        
!  elements from the matrix A to get the pattern. when omega -->0,      
!  the pattern will --> pattern of A while when omega --> 1 the         
!  pattern will --> diagonal because all diagonal elements shall be     
!  in the pattern. When omega >1.0 then the pattern was determined      
!  by the symbolic factorization for ILU(s), when s is set to be        
!  int(omega).                              
!                                   
!   (2) Space allocation for ECIMGS with the given sparsity     
!  pattern. The difference from the symbolic factorization of ILU
!  is that, at each step i, if (i,j) is in the target pattern of the    
!  factors, then all the elements in jth row will be considered in  
!  fillins. while if (i,j) is only a needed extra space, then only  
!  the elements in target pattern will be consider to fill into row i   
!                                   
!   The implementation assumes that the diagonal entries are    
!  nonzero, and remain nonzero throughout the elimination       
!  process.  The algorithm proceeds row by row.  When computing     
!  the sparsity pattern of the i-th row, the effect of row      
!  operations from previous rows is considered.  Only those     
!  preceding rows j for which (i,j) is nonzero need be considered,  
!  since otherwise we would not have formed a linear combination    
!  of rows i and j.                         
!                                                                       
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
!
! on entry:
!========== 
!  n       = The order of the matrix A.
!  A, c_ind, r_point = The structure of matrix A stored in 
!                CSR format.
!  omega   = real parameter for pattern selection.
!

! on return:
!=========== 
!
! ilu   = The integer pointer. iwork(ilu) is the array of ILU preconditioners.
! ilup   = Integer array of length n containing the pointers to        
!           the beginning of each row of u in the matrix ldu,ijlu. 
! ierr   = Error message with the following meaning.
!           ierr  =  0   --> successful return.
!           ierr  = -1   --> not enough storage, mneed contains the
!                        approximate additional storage needed for ldu
!
! mneed   = contains the actual number of elements in ilu, or the amount
!       of additional storage needed for ilu
!
! work arrays:
!=============
! iwork, space  = whole working space. All the working arrays are subarray
!          of this array ( size = space)
! rowll      = index to the integer array of length n containing pointers 
!          to implement a linked list for the fill-in elements.
!
! calling routines:
!==================
!
! external functions:
!====================
! int
!
! note:
!======
! the columns of matrix A shall be sorted in increase order.
!-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-* 
!
!----------------------------------------------------------------------- 
      integer n,c_ind(*), r_point(*)
      integer ilu,ilup,ilup1, ilu1, rowll
      integer iwork(*), space, size_work
      integer ierr,   mneed,  used
      integer i,j,k,ij,lp,m
      integer prev,lm,k1
      double precision ::  A(*), omega
!
!  determine the sparsity pattern.
!
    size_work = space
    if ( omega .eq. 0.0d0) then
!   pattern is equal to pattern of A
        call denser(n, c_ind, r_point,   omega,         &
                         ilu, ilup,  iwork, size_work,  &
                         ierr, mneed )
        if (ierr.ne.0) then
!       print*, 'error in pattern selection case one'
        goto 1000
        endif
    elseif (omega.gt.0.0d0 .and. omega.lt.1.0d0) then
!   pattern is sparser then pattern A by dropping smaller elements
        call sparser(n, A, c_ind , r_point,   omega,            &
                         ilu   ,ilup  ,  iwork , size_work ,    &
                         ierr  , mneed )
        if (ierr.ne.0) then
!                print*, 'error in pattern selection case two'
                goto 1000
            endif
    elseif ( omega.eq.1.0d0) then
!   pattern equal to diagonal
        call diagonal(n, c_ind , r_point,                       &
                         ilu   ,ilup  ,  iwork , size_work  ,   &
                         ierr  , mneed )
            if (ierr.ne.0) then
!                print*, 'error in pattern selection case three'
                goto 1000
            endif
    elseif (omega.gt.1.0d0) then
!   pattern is denser than A (allow fillins)
        call denser(n     ,c_ind , r_point,   omega,            &
                         ilu   ,ilup  ,  iwork , size_work ,    &
                         ierr  , mneed )
            if (ierr.ne.0) then
!                print*, 'error in pattern selection case four'
                goto 1000
            endif
    endif
!
!  Initialize the working arrays for symbolic analysis.
!
    used  = ilu + iwork(ilu+n) - 1
        rowll = used + 1
        rowll = rowll + mod(rowll-1, intdbl)
        ilup1 = rowll + n + 1
        ilup1 = ilup1 + mod(ilup1-1, intdbl)
        ilu1  = ilup1 + n
        ilu1 = ilu1 + mod(ilu1-1, intdbl)
!
!  Initialize the working array rowll
!
        do i = 1,n+1
            iwork(rowll+i-1) = 0
        enddo
!     
!  Because the first row of the factor contains no strictly lower
!  triangular parts (parts of L), ijlu(1) = n+2:
!
        iwork(ilu1)  =  n+2
    lp = n+2
!
! Proceed row by row:
!
        do i = 1,n
!
!  Get initial structure of row i
!
        iwork(rowll) = n+1
        k1 = 0
            do k = iwork(ilu+i-1), iwork(ilu+i)-1
            j = iwork(ilu+k-1)
            iwork (rowll+abs(j)) = iwork(rowll+k1)
            iwork (rowll+k1) = j
            k1 = abs(j)
        enddo
!
!  Compute fillins
!
        j = iwork(rowll)
        do while ( abs(j).lt.i) 
            if (j.gt.0) then
!               only fillin R elements
                    prev = j
                    lm = abs(iwork(rowll+j))
                    do m = iwork(ilup1+j-1), iwork(ilu1+j)-1
                        ij = iwork(ilu1+m-1)
                        if (ij.lt.0) then
                            ij = abs(ij)
                            do while (lm .le. ij)
                                prev = lm
                                lm   = abs(iwork(rowll+lm))
                            enddo
                            if (prev.lt.ij) then
!                               insert element
                                iwork(rowll+ij) = iwork(rowll+prev)
                                iwork(rowll+prev) = ij
                                prev = ij
                            endif
                        endif
                    enddo
            elseif (j.lt.0) then
!                   fillin all elents of P
                j = abs(j)
                    prev = j
                    lm = abs(iwork(rowll+j))
                    do m = iwork(ilup1+j-1), iwork(ilu1+j)-1
                        ij = abs(iwork(ilu1+m-1))
                        do while (lm .le. ij)
                            prev = lm
                            lm   = abs(iwork(rowll+lm))
                        enddo
                        if (prev.lt.ij) then
!                           insert element
                            iwork(rowll+ij) = iwork(rowll+prev)
                            iwork(rowll+prev) = ij
                            prev = ij
                        endif
                    enddo
                endif
            j = iwork(rowll+j)
            enddo
!
! set row i 
!
            lm = iwork(rowll)
            do while (abs(lm) .lt. i)
                iwork(ilu1+lp-1) = lm
                lp = lp + 1
        if (ilu1+lp .gt.space) then
!           short of space
            ierr = 1
            mneed = ifix((float(n-i)/float(2*i))*3*lp)
            goto 1000
        endif
                lm = iwork(rowll+abs(lm))
            enddo
        iwork(ilup1+i-1) = lp
        if (abs(lm).eq.i) lm = iwork(rowll+abs(lm))
        do while (abs(lm) .le.n) 
        iwork(ilu1+lp-1) = lm
        lp = lp + 1
        if (ilu1+lp .gt.space) then
!           short of space
            ierr = 1
            mneed = ifix((float(n-i)/float(2*i))*3*lp)
            goto 1000
        endif
        lm = iwork(rowll+abs(lm))
        enddo
        iwork(ilu1+i) = lp
        enddo
    mneed = max( mneed, ilu1+lp-1)
!
! Collect the working space
!
    do i = ilup1, ilu1+lp
        iwork(i-ilup1+1) = iwork(i)
    enddo
    ilup = 1
    ilu = n+1
    ilu = ilu + mod(ilu-1,intdbl)

1000    continue
        return
        end


