!========================================================================
      subroutine tridsl( trans, n, dl, d, du, du2, ipiv, b, info )


!**********************************************************************
!                   SPLIB Copyright (C) 1995-1999                    **
!                     Indiana University                             **
!**********************************************************************

!========================================================================
!      ***** Solve a factored tridiagonal system *****          *
!========================================================================
!
!     Solve a linear system Ax = b, where A is tridiagonal and has
!     been factored by trifac.  This routine is modified from a Lapack
!     routine (dgttrs.f).  Modifications include changing the code to
!     handle only one right hand side, and eliminating xerbla calls.
!     Because of the change in the calling sequence, the error codes
!     have also been changed.
!
!------------------------------------------------------------------------
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Tue Jan 18 10:01:23 EST 1994
!------------------------------------------------------------------------
!
!  -- original Lapack routine (Version 1.1) --
!     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
!     Courant Institute, Argonne National Lab, and Rice University
!     March 31, 1993
!
!
!  Scalar arguments 
      implicit none
      integer            info, n
!  Array arguments 
      integer            ipiv( * )
      double precision   b( * ), d( * ), dl( * ), du( * ), du2( * )
!  Character arguments 
      character trans(*)
!
!========================================================================
!
!  purpose
!  =======
!
!  tridsl solves one of the systems of equations
!     a*x = b  or  a'*x = b,
!  with a tridiagonal matrix a using the lu factorization computed
!  by trifac.f.
!
!  arguments
!  =========
!
!  trans   (input) character
!          specifies the form of the system of equations:
!          = 'n':  a * x = b  (no transpose)
!          = 't':  a'* x = b  (transpose)
!
!  n       (input) integer
!          the order of the matrix a.
!
!  dl      (input) double precision array, dimension (n-1)
!          the (n-1) multipliers that define the matrix l from the
!          lu factorization of a.
!
!  d       (input) double precision array, dimension (n)
!          the n diagonal elements of the upper triangular matrix u from
!          the lu factorization of a.
!
!  du      (input) double precision array, dimension (n-1)
!          the (n-1) elements of the first superdiagonal of u.
!
!  du2     (input) double precision array, dimension (n-2)
!          the (n-2) elements of the second superdiagonal of u.
!
!  ipiv    (input) integer array, dimension (n)
!          the pivot indices; for 1 <= i <= n, row i of the matrix was
!          interchanged with row ipiv(i).  ipiv(i) will always be either
!          i or i+1; ipiv(i) = i indicates a row interchange was not
!          required.
!
!  b       (input/output) double precision array, dimension (n)
!          on entry, the right hand side matrix b.
!          on exit, b is overwritten by the solution matrix x.
!
!  info    (output) integer
!          = 0:  successful exit
!          < 0:  if info = -i, some argument had an illegal value
!
!========================================================================
!========================================================================
!
!  Local scalars: 
      logical            notran
      integer            i
      double precision   temp
!  External functions: 
      logical            lsame
      external           lsame
!  Intrinsic functions:
      intrinsic          max
!
!========================================================================
!
      info = 0
      notran = trans(1) .eq. 'n'  .or. trans(1) .eq. 'N'
      if( n.lt.0 ) then
         info = -2
      end if
      if( info.ne.0 ) then
         write(6,*)'Error in tridsl; info = ',info
         return
      end if
      if( n.eq.0) return
!
      if( notran ) then
!
!        solve a*x = b using the lu factorization of a,
!        overwriting right hand side vector with its solution.
!
!           solve l*x = b.
!
            do 10 i = 1, n - 1
               if( ipiv( i ).eq.i ) then
                  b( i+1) = b( i+1 ) - dl( i )*b( i )
               else
                  temp = b( i )
                  b( i ) = b( i+1 )
                  b( i+1 ) = temp - dl( i )*b( i )
               end if
   10       continue
!
!           solve u*x = b.
!
            b( n ) = b( n ) / d( n )
            if( n.gt.1 ) b( n-1 ) = ( b( n-1 )-du( n-1 )*b( n ) ) / d( n-1 )
            do 20 i = n - 2, 1, -1
               b( i ) = ( b( i )-du( i )*b( i+1 )-du2( i )*b( i+2 ) ) / d( i )
   20       continue
      else
!
!        solve a' * x = b.
!
!
!           solve u'*x = b.
!
            b( 1 ) = b( 1 ) / d( 1 )
            if ( n.gt.1 ) b( 2 ) = ( b( 2 )-du( 1 )*b( 1 ) ) / d( 2 )
            do 40 i = 3, n
               b( i ) = ( b( i )-du( i-1 )*b( i-1 )-du2( i-2 )*b( i-2 ) ) / d( i )
   40       continue
!
!           solve l'*x = b.
!
            do 50 i = n - 1, 1, -1
               if( ipiv( i ).eq.i ) then
                  b( i ) = b( i ) - dl( i )*b( i+1 )
               else
                  temp = b( i+1 )
                  b( i+1 ) = b( i ) - dl( i )*temp
                  b( i ) = temp
               end if
   50       continue
      end if
!
!  End of tridsl
!
      return
      end
