!=========================================================================
      subroutine  bicgs (A,colind,rwptr,n,                  &
                         x,r,b,                             &
                         tolrnc,reltv,maxits,ierr,          &
                         work,wksz,                         &
                         lu,jlu,uptr,pcmeth,iunit,instlvl)

!-----------------------------------------------------------------------
!                           *** Bi-CGSTAB ***				
!									
!  Van der Vorst's biconjugate gradients stabilized algorithm.		
!  This version is a direct translation of the Templates version to	
!  the SPLIB paradigm.  Because it uses a different change of variables	
!  to handle the preconditioning, its performance and convergence	
!  can differ greatly from the CGSTAB routine of SPLIB, which in	
!  theory implements the same algorithm.  Double (v = v + a*w + b*u)	
!  vector updates are done in a single loop.  For machines like the	
!  IBM RS/6000, this gives a more accurate update than using a sequence	
!  of scalings and vector updates because the combined multiply/add	
!  unit retains the full result of a multiply (106 bit mantissa).	
!									
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array of additional storage.  Bicg
!            requires 6 vectors of length n, and if instrumentation
!            routine instr.f is called, an additional 2 vectors of
!            length n are required.  If not enough storage is given,
!            nothing is done and an error code is returned specifying
!            the additional storage needed (in double precision words).
!  wksz    : Length of array work, in double precision words.  
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL.  Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tolsq is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tolsq is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by bicgs, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  early   : Logical parameter that indicates if a tentative exit
!            was made because of early termination.  This is needed
!            because of the way SPLIB uses unpreconditioned residuals
!            as the stopping test.
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  urnorm  : Unpreconditioned initial residual.  Used for tests to see
!            if initial residual is good enough, and if preconditioner
!            is no good (see pdr above).
!  Mrnorm  : preconditioned initial residual.
!
!  rtld,t,
!  p,ptld,
!  stld,v  : Integer pointers in the array work(*) that give the
!            starting position of the corresponding vector used by BICGS.
!            Note that the array s in Templates bicgs is identified
!            with r in this routine.
!  wins1,
!  wins2   : Integer pointers in the array work(*) that give starting
!            positions of two vectors used by instrumentation routines.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!           ierr = 5  --> error condition of abs(omega) <= SMALL.
!           ierr = 6  --> error condition of abs(rho) <= SMALL.
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, daxpy, dcopy
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Mon Jul 25 12:39:12 EST 1994
!     Documented Tue Jul  4 15:19:35 EST 1995
!     Approved by Radha, Tue Jul  4 16:20:57 EST 1995
!
!----------------------------------------------------------------------
       implicit none

!     -------------------------
!     Declaration of arguments:
!     -------------------------
       integer n
       double precision :: A(*)
       double precision :: x(*), b(*), r(*)
       double precision :: work(n,*)
       double precision :: tolrnc(*)
       double precision :: lu(*)

       integer colind(*), rwptr(*)
       integer wksz
       integer maxits, ierr
       integer jlu(*), uptr(*)
       integer pcmeth, iunit(*), instlvl
       logical reltv

!     ------------------
!     External functions
!     ------------------
      double precision :: ddot, dnrm2 
      external ddot, dnrm2

!     ------------- 
!     Local storage 
!     ------------- 
      logical first, early
      integer i, its, resets
      double precision :: alpha, beta, rho, rho1, omega, tolsq,temp, res
      double precision :: SMALL, pdr, Mrnorm, rnorm

!     -------------------------------
!     Pointers for work array vectors
!     -------------------------------
      integer rtld, p, v, t, ptld, stld, wins1, wins2

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      pdr = 1.0d0/epsilon(SMALL)

!----------------------------------------------------------------------
!
!  Key to vectors in array work
!
!       s    -> r          (not used here; r is just substituted)
!       rtld -> work(:,1)  (arbitrary fixed vector)
!       p    -> work(:,2)
!       v    -> work(:,3)
!       t    -> work(:,4)
!       ptld -> work(:,5)
!       stld -> work(:,6)
!
!----------------------------------------------------------------------

!     -------------------------------------
!     Set up pointers to vectors in work(*)
!     -------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
      rtld = wins2 + 1
      p    = rtld + 1
      v    = p + 1
      t    = v + 1
      ptld = t + 1
      stld = ptld + 1

!     ==========================
!     Initialize for iterations:
!     ==========================

      first = .true.
      resets = 0
      tolsq = tolrnc(1)*tolrnc(1)

!     ----------------------------------------
!     Return if not enough workspace provided:
!     ----------------------------------------
      if (wksz .lt. (wins2+6)*n) then
         ierr = wksz - (wins2+6)*n
         maxits = 0
         return
      end if 

!     --------------------------------------------
!     Indication of early termination in middle of
!     computing two vectors for updating x:
!     --------------------------------------------
      early = .false.
      its = 0

!     ---------------------------
!     Initial residual r = b - Ax
!     ---------------------------
      call bmux(n,n,x,r,A,colind,rwptr)
      do i = 1,n
         r(i) = b(i) - r(i)
      end do
      rnorm = ddot(n,r,1,r,1)

!     -------------------------------------------------
!     Adjust tolerance if relative residual to be used:
!     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*rnorm
          tolrnc(1) = tolrnc(1)*sqrt(rnorm)
      end if

!     -----------------------------------------
!     Stop if initial residual is small enough:
!     -----------------------------------------
      if (rnorm .le. tolsq) go to 900

!     ------------------------------------------
!     Note: r-tilde can be set arbitrarily here.
!     ------------------------------------------
      call dcopy( n, r, 1, work(1,rtld), 1 )


!=======================================================================
!     Begin main loop of iterations.  Loop is equivalent to
!      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
!=======================================================================

 20      continue

         call instr(A,colind,rwptr,n, x,b,work(1,wins1), lu,jlu,uptr,   &
                    pcmeth,first,its, iunit,instlvl)
         its = its + 1
         rho = ddot(n,work(1,rtld),1,r,1)
         if (abs(rho) .le. SMALL) go to 930

!        ----------------------------
!        p = beta*(p - omega*v) + r
!        (on first iteration, p = r).
!        ----------------------------
         if (its .gt. 1) then
            beta = (rho/rho1)* (alpha/omega)
            do i = 1, n
               work(i,p) = r(i) + beta*(work(i,p) - omega*work(i,v))
            end do
         else
            call dcopy( n, r, 1, work(1,p), 1 )
         end if

!        ----------------
!        Solve M*ptld = p
!        ----------------
         call lusl(n,work(1,p),work(1,ptld),lu,jlu,uptr,pcmeth,     &
                   A,colind,rwptr)

!        ------------------------------------------
!        Check initial preconditioned residual size 
!        ------------------------------------------
         if (its .eq. 1) then
            Mrnorm = ddot(n, work(1,ptld), 1, work(1,ptld), 1)
            if (Mrnorm .gt. pdr*pdr*rnorm) then
               ierr = 4
               maxits = 0
               return
            end if 
         end if 

!        --------------
!        Set v = A*ptld
!        --------------
         call bmux(n, n,work(1,ptld),work(1,v),A,colind,rwptr)

!        ------------------
!        alpha = rho/v'rtld
!        ------------------
         alpha = rho/ddot( n, work(1,rtld), 1, work(1,v), 1 )

!        ------------------------------------------------
!        s = r - alpha*v.  Note s and r are aliased here.
!        ------------------------------------------------
         do i = 1, n
            r(i) = r(i) - alpha*work(i,v)
         end do

!        ---------------------------------------
!        Early termination if s is small enough:
!        ---------------------------------------
         if ( ddot(n,r,1,r,1) .le. tolsq) then

!           -------------------------------------------------------------
!           Indicate early termination in case true res not small enough.
!           -------------------------------------------------------------
            early = .true.

!           -------- 
!           Update x 
!           -------- 
            call daxpy( n, alpha, work(1,ptld), 1, x, 1 )
            go to 910
         end if 
 30      continue

!        -----------------------------------------
!        Compute stabilizer vector stld = inv(M)*s
!        -----------------------------------------
         call lusl(n,r,work(1,stld),lu,jlu,uptr,pcmeth,A,colind,rwptr)

!        ------------------
!        Compute t = A*stld
!        ------------------
         call bmux(n, n,work(1,stld),work(1,t),A,colind,rwptr)
         omega = ddot( n, work(1,t), 1, r, 1 ) /            &
                 ddot( n, work(1,t), 1, work(1,t), 1 )

!        --------
!        Update x
!        --------
         do i = 1, n
            x(i) = x(i) + alpha*work(i,ptld) + omega*work(i,stld)
         end do

!        --------------------------------------
!        Update residual vector r = s - omega*t
!        --------------------------------------
         call daxpy( n, -omega, work(1,t), 1, r, 1 )
         res = ddot(n,r,1,r,1)
         if (res .le. tolsq) go to 910
         if (its .ge. maxits) go to 920
         if (abs(omega) .le. SMALL) go to 940
         rho1 = rho
         go to 20

!=======================================================================
!                  Stopping conditions
!=======================================================================

!     ====================
!     (Almost) exact guess
!     ====================
 900  continue
      ierr = 1
      maxits = 0
      return

!     ========================================= 
!     Possible convergence; check true residual 
!     ========================================= 
 910  continue
      ierr = 0

!     Stop if residual tolerance reset 8 times.
!     -----------------------------------------
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n,n,x,work(1,ptld),A,colind,rwptr)
      do i = 1, n
         work(i,ptld) = b(i) - work(i,ptld)
      end do
      res = ddot(n,work(1,ptld),1,work(1,ptld),1)
      temp = tolrnc(1)*tolrnc(1)
      if (res .gt. temp) then
          tolsq = tolsq*temp/res
!          write(6,*)'>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          if (early) then
             early = .false.
             alpha = 0.0d0
             go to 30
          end if
          go to 20
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     ========================================
!     No convergence in max allowed iterations
!     ========================================
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     =============
!     rho too small
!     =============
 930  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     ===============
!     omega too small
!     ===============
 940  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!================= End of BICGS =======================================

      end
