!========================================================================
      subroutine  cgne(A,colind,rwptr,n,                    &
                       x,r,b,                               &
                       tolrnc,reltv,maxits,ierr,            &
                       work,wksz,                           &
                       lu,jlu,uptr,pcmeth,iunit,instlvl)

!-----------------------------------------------------------------------
!                   *** CGNE ***                    
!                                   
!  Method originally proposed in Hestenes and Stiefel's classical   
!  paper on the CG method.                      
!                                   
!  Solve a nonsymmetric sparse system Ax = b by the conjugate gradient  
!  iteration applied to the normal equations,               
!                                   
!                        AA'y = b     x = A'y               
!                                   
!  The sparse matrix is stored in the compressed sparse row (CSR)   
!  format (a, colind, rwptr).                       
!                                   
!  This uses two sided preconditioning, implicitly working on the   
!  linear system                            
!                                   
!     inv(LU)*A*A'*inv(LU)' v = inv(LU)*b,              
!                           x = A'*inv(LU)'             
!                                   
!  The algorithm does not directly use v, and instead via a change of   
!  variables it directly updates the vector x.  This two sided version  
!  uses the changes of variables                    
!   d = inv(LU)'*dhat, where dhat is the usual search direction,    
!   q = A'*d,                           
!   x = A'*y = A'* inv(LU)'*v,                  
!   y = inv(LU)'*v.                         
!  The residual vector r actually being kept here is for the residual   
!      r = inv(LU)*(b-A*x).                     
!                                   
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array of additional storage.  CGNE
!            requires 3 vectors of length n, and if instrumentation
!            routine instr.f is called, an additional 2 vectors of
!            length n are required.  If not enough storage is given,
!            nothing is done and an error code is returned specifying
!            the additional storage needed (in double precision words).
!  wksz    : Length of array work, in double precision words.  
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL.  Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tolsq is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tolsq is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by cgne, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  urnorm  : Unpreconditioned initial residual.  Used for tests to see
!            if initial residual is good enough, and if preconditioner
!            is no good (see pdr above).
!
!  w,d,q   : Integer pointers in the array work(*) that give the
!            starting position of the corresponding vector used by CGNE.
!  wins1,
!  wins2   : Integer pointers in the array work(*) that give starting
!            positions of two vectors used by instrumentation routines.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!           ierr = 5  --> error condition for bs(qtq) <= SMALL.
!           ierr = 6  --> error condition for abs(rho) <= SMALL.
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  utltsl  : Solves M'z = r.
!  bmux    : Computes matrix-vector product with A.
!  btmux   : Computes matrix-vector product with A'.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, daxpy
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Thu Jul 28 13:40:51 EST 1994
!     Documented Wed Jul  5 12:56:48 EST 1995
!
!----------------------------------------------------------------------
       implicit none

!     -------------------------
!     Declaration of arguments:
!     -------------------------
      integer n
      double precision :: A(*)
      double precision :: x(*), b(*), r(*)
      double precision :: work(n,*)
      double precision :: tolrnc(*)
      double precision :: lu(*)
!
      integer colind(*), rwptr(*)
      integer jlu(*), uptr(*)
      integer maxits, ierr, iunit(*), instlvl
      integer pcmeth, wksz
      logical reltv

!     ------------------
!     External functions
!     ------------------
      double precision :: ddot

!     -------------
!     Local storage 
!     ------------- 
      double precision :: urnorm, pdr, SMALL, t
      integer its, i, resets
      double precision :: alpha, beta, temp, rho, tolsq
      logical first

!     -------------------------------
!     Pointers for work array vectors
!     -------------------------------
      integer w, d, q, wins1, wins2

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      pdr = 1.0d0/epsilon(SMALL)

!----------------------------------------------------------------------
!  Key to work storage:
!     w -> work(:,1)  (product vector w = A*p, etc.)
!     d -> work(:,2)  (search direction d)
!     q -> work(:,3)  (preconditioned residual vector)
!----------------------------------------------------------------------

!     --------------------------------------
!     Set up pointers for work array vectors
!     --------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
      w = wins2 + 1
      d = w + 1
      q = d + 1

!     -------------------------
!     Initialize for iterations
!     -------------------------
      its = 0
      resets = 0
      tolsq = tolrnc(1)*tolrnc(1)
      first = .true.

!     ----------------------------------------
!     Return if not enough workspace provided:
!     ----------------------------------------
      if (wksz .lt. (wins2+3)*n) then
         ierr = wksz - (wins2+3)*n
         maxits = 0
         return
      end if 

!     --------------------------------------
!     Check initial true residual w = b - Ax 
!     --------------------------------------
      call bmux(n, n, x, work(1,w), A, colind, rwptr)
      do 10 i = 1, n
         work(i,w) = b(i) - work(i,w)
  10  continue
      urnorm = ddot(n, work(1,w), 1, work(1,w), 1)

!     ------------------------------------------------
!     Adjust tolerance if relative residual to be used
!     ------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

!     -----------------------------------------------------
!     Check for initial guess satisfying stopping condition
!     -----------------------------------------------------
      if (urnorm .le. tolsq) go to 900

!     ------------------------------
!     r = inv(LU)*w = inv(LU)*(b-Ax)
!     ------------------------------
      call lusl(n, work(1,w), r, lu, jlu, uptr, pcmeth, A,colind,rwptr)

!     ---------
!     rho = r'r
!     ---------
      rho = ddot(n, r, 1, r, 1)

!     --------------
!     d = inv(LU)'*r
!     --------------
      call utltsl(n, r, work(1,d), lu, jlu, uptr, pcmeth,       &
                  A,colind,rwptr)

!  ---------------------------------------------------------
!  Check if preconditioned residual is too large relative to
!  unpreconditioned residual:
!  ---------------------------------------------------------
      if (ddot(n,work(1,d),1,work(1,d),1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

!=======================================================================
!     Begin main loop of iterations.  Loop is equivalent to
!      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
!=======================================================================

  30   continue

      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      its = its + 1

!     -------
!     q = A'd
!     -------
      call btmux(n,n, work(1,d), work(1,q), A, colind, rwptr)

!     ---------------
!     w = inv(LU)*A*q
!     ---------------
      call bmux(n, n, work(1,q), work(1,w), A, colind, rwptr)
      call lusl(n, work(1,w), work(1,w), lu,jlu,uptr, pcmeth, A,    &
                colind,rwptr)

!     ---------------
!     alpha = rho/q'q
!     ---------------
      t = ddot(n, work(1,q), 1, work(1,q), 1)
      if (abs(t) .le. SMALL) go to 930
      alpha = rho / t

!     ---------------
!     x = x + alpha*q
!     ---------------
      call daxpy(n, alpha, work(1,q), 1, x, 1)

!     ---------------
!     r = r - alpha*w
!     ---------------
      call daxpy(n, -alpha, work(1,w), 1, r, 1)

!     ---------------------------------------
!     temp = r'r, beta = temp/rho, rho = temp
!     ---------------------------------------
      temp = ddot(n, r, 1, r, 1)
      if (abs(rho) .le. SMALL) go to 940
      beta = temp / rho
      rho = temp

!     -----------------------
!     d = inv(LU)'*r + beta*d
!     -----------------------
      call utltsl(n, r, work(1,w), lu, jlu, uptr, pcmeth, A,    &
                  colind,rwptr)
      do 50 i = 1,n
                work(i,d) = work(i,w) + beta*work(i,d)
  50  continue
!
      if (rho .le. tolsq) go to 910
  60  continue
      if (its .gt. maxits) go to 920
      go to 30

!=======================================================================
!                  Stopping conditions
!=======================================================================

!     ====================
!     (Almost) exact guess
!     ====================
 900  continue
      ierr = 1
      maxits = 0
      return

!     =========================================
!     Possible convergence; check true residual
!     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n, n,x,work(1,w),A,colind,rwptr)
      do i = 1, n
         work(i,w) = b(i) - work(i,w)
      end do
      temp = ddot(n,work(1,w),1,work(1,w),1)
      if (temp .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/temp
!          write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 60
      end if
 915  continue
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     =========================================
!     No convergence in max allowed iterations.
!     =========================================
 920  continue
      ierr = 3
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     ================================
!     Failure from abs(q'q) too small.
!     ================================
 930  continue
      ierr = 5
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!     ================================
!     Failure from abs(rho) too small.
!     ================================
 940  continue
      ierr = 6
      maxits = its
      call instr(A,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its, iunit,instlvl)
      return

!================= End of CGNE ======================================== 

      end
