!=======================================================================
      subroutine gmrest(A,colind,rwptr,n,                       &
                        x,r,b,                                  &
                        tolrnc,reltv,maxits,ierr,               &
                        work,wksz,                              &
                        lu,jlu,uptr,pcmeth,iunit,instlvl)

!-----------------------------------------------------------------------
!                                                                       
!                 *** GMRES (Templates version) ***         
!                                                                       
!  This is the version of GMRES(m) from Templates, changed to fit in    
!  the paradigm used for SPLIB.  It is provided to allow comparisons    
!  with the native SPLIB version of GMRES(m), and differs in its    
!  implementation primarily in how the orthogonal rotations are     
!  computed and applied, and in using left preconditioning instead  
!  of right preconditioning.  The reverse communication model       
!  has been replaced with SPLIB's forward communications.  This     
!  routine is just a wrapper, to divide up the work() array and then    
!  call the slightly modified Templates routine, renamed gmres1 here.   
!                                                                       
!-----------------------------------------------------------------------
!                                                                       
!     Randall Bramley                           
!     Department of Computer Science                    
!     Indiana University                        
!     email: bramley@cs.indiana.edu                 
!     Tue Aug  2 16:53:53 EST 1994                  
!     Documented Mon Jul 10 10:15:28 EST 1995               
!                                                                       
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : vector of unknowns to be solved for.
!  b       : right hand side vector of system to be solved: Ax = b
!  r       : residual vector.
!  work    : work array 
!  wksz    : size of work array 
!
!     (Preconditioner data structure)
!  lu,jlu  : matrix stored in Modified Sparse Row (MSR) format containing
!            the L and U factors together. The diagonal (stored in
!            lu(1:n) ) is inverted. Each i-th row of the lu,jlu matrix
!            contains the i-th row of L (excluding the diagonal entry=1)
!            followed by the i-th row of U.
!  uptr    : integer array of length n containing the pointers to
!            the beginning of each row of U in the matrix lu,jlu.
!  pcmeth  : Preconditioning method to use; see splib.f.
!
!     (Iteration control)
!  tolrnc  : tolerances on norm(s) used in stopping test.
!  maxits  : Maximum number of iterations allowed; also maximal size
!            of Krylov subspace to use.
!            maxits(1) ---> maximum iterations to allow.
!            maxits(2) ---> maximal Krylov subspace size.
!
!      (Other stuff)
!  ierr    : Error flag; see return values below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by gmrest, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!   Local variables
!   ---------------
!  h       : integer pointer for begin position of the Hessenberg
!            matrix H in work array
!  w       : integer pointer for begin position of the Krylov subspace 
!            basis vectors in work
!  wkins   : integer pointer for begin position of the work array for 
!            instrumentation routine
!  pdr     : parameter to determine if preconditioner is unstable.
!  urnorm  : unpreconditioned initial residual.
!  im      : size of Krylov subspace.
!
!----------------------------------------------------------------------
!
!
! Return values:
! --------------
!
!  x      : Contains an approximate solution (upon successful return).
!
!  ierr   : Error flag with the following meanings:
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!
!----------------------------------------------------------------------
!  Routines
!  --------
!  gmres1  : Modified Templates version of gmres
!  bmux    : Multiplies y = A*x
!  lusl    : Solves M z = r.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS1 routines: ddot, daxpy, dnrm2, dscal 
!  BLAS2 routine: dgemv
!----------------------------------------------------------------------* 
        implicit none
        integer n, im, maxits(2), ierr
        double precision :: work(*), b(n), x(n), A(*), lu(*), r(*)
        double precision :: tolrnc(*)
        integer colind(*), rwptr(n+1), jlu(*), uptr(n)
        integer pcmeth, instlvl, iunit(*)
        integer wksz
        integer kwork
        integer h, wkins, w
        logical reltv

!       ------------------------------------
!       Unpack maximal Krylov subspace size.
!       ------------------------------------
        im = maxits(2)

!       ----------------------------------------------------
!       Set pointers and check if enough workspace provided.
!       ----------------------------------------------------
        wkins =  1
        if (instlvl  >  0) then
           w = wkins + 2*n
        else
           w = wkins
        end if
        h     = w + n*(im+4)
        kwork = h + (im+1)*(im+2)

        if (wksz  <  kwork) then
           ierr = wksz - kwork
           maxits(1) = 0
           return
        end if
!
!       ---------------------------------------------
!       Call the gmres1.f routine (Templates version)          
!       ---------------------------------------------
        call gmres1(A,colind,rwptr,n,                       &
                    x,r,b,                                  &
                    tolrnc(1),reltv,maxits(1),im,ierr,      &
                    work(w),n,work(h),im+1,                 &
                    lu,jlu,uptr,pcmeth,work(wkins),iunit, instlvl)

        return
        end subroutine gmrest

!======================================================================

      subroutine gmres1(A,colind,rwptr,n,                   &
                        x,res,b,                            &
                        tolrnc,reltv,maxits,restrt,ierr,    &
                        work,ldw,h,ldh,                     &
                        lu,jlu,uptr,pcmeth,wkins,iunit, instlvl)

       implicit none
       integer n, ldw, ldh
       double precision :: work(ldw,*), b(n), x(n), A(*), lu(*), res(*)
       double precision :: wkins(*)
       double precision :: tolrnc
       integer restrt, maxits, ierr
       integer colind(*), rwptr(n+1), jlu(*), uptr(n)
       integer pcmeth, iter, resets, instlvl, iunit(*)
       logical first, reltv
       double precision :: h( ldh,* )
       double precision :: pdr, urnorm
!=======================================================================
!  
! WARNING: the below documentation is from the original TEMPLATES
! version of this, and has been changed.  In particular, the matrix
! and its preconditioner are explicitly passed in, and the "reverse
! communication" model of Templates has been changed to the "forward
! communication" model of SPLIB.  This has required changes in the
! calls to the matrix-vector and preconditioning routines.  Other
! changes include the error flag modifications, and calls to the
! instrumentation routines.
!
! Although credit is given below to the Templates authors, any errors
! in this version probably should be attributed to the modifications
! made; follow the instructions below to get the Templates documentation
! and information for getting an unmodified version from netlib.
!
!
!  -- Iterative template routine --
!     Univ. of Tennessee and Oak Ridge National Laboratory
!     October 1, 1993
!     Details of this algorithm are described in "Templates for the
!     Solution of Linear Systems: Building Blocks for Iterative
!     Methods", Barrett, Berry, Chan, Demmel, Donato, Dongarra,
!     Eijkhout, Pozo, Romine, and van der Vorst, SIAM Publications,
!     1993. (ftp netlib2.cs.utk.edu; cd linalg; get templates.ps).
!
!  purpose
!  =======
!
!  gmres solves the linear system ax = b using the
!  generalized minimal residual iterative method with preconditioning.
!
!  convergence test: ( norm( b - a*x ) / norm( b ) ) < tolrnc.
!  for other measures, see the above reference.
!  (Note: convergence test is norm( b - a*x ) < tolrnc if reltv = .false.)
!
!  arguments
!  =========
!
!  n       (input) integer. 
!  on entry, the dimension of the matrix.
!  unchanged on exit.
! 
!  b       (input) double precision array, dimension n.
!  on entry, right hand side vector b.
!  unchanged on exit.
!
!  x       (input/output) double precision array, dimension n.
!  on input, the initial guess; on exit, the iterated solution.
!
!  restrt  (input) integer
!  restart parameter, <= n. this parameter controls the amount
!  of memory required for matrix h (see work and h).
!
!  work    (workspace) double precision array, dimension (ldw,restrt+4).
!
!  ldw     (input) integer
!  the leading dimension of the array work. ldw >= max(1,n).
!
!  h       (workspace) double precision array, dimension (ldh,restrt+2).
!  this workspace is used for constructing and storing the
!  upper hessenberg matrix. the two extra columns are used to
!  store the givens rotation matrices.
!
!  ldh    (input) integer
!  the leading dimension of the array h. ldh >= max(1,restrt+1).
!
!  iter    (input/output) integer
!  on input, the maximum iterations to be performed.
!  on output, actual number of iterations performed.
!
!  resid   (input/output) double precision
!  on input, the allowable convergence measure for
!  norm( b - a*x ) / norm( b ).
!  on output, the final value of this measure.
!
!  matvec  (external subroutine)
!  lusol (external subroutine)
! (Note: matvec, lusol have been replaced with bmux, lusl)
!
!  info    (output) integer
!  =  0: successful exit. iterated approximate solution returned.
!  >  0: convergence to tolerance not achieved. this will be
!        set to the number of iterations performed.
!  <  0: illegal input parameter.
!           -1: matrix dimension n < 0
!           -2: ldw < n
!           -3: maximum number of iterations iter <= 0.
!           -4: ldh < restrt
! (Note: info has been replaced with ierr, with different 
!  return values.  See gmrest.f above).
!
!  blas calls:   daxpy, dcopy, ddot, dnrm2, drot, drotg, dscal
!  ============================================================
!
      integer   i, k, av, cs, sn, r, s, v, w, y
      double precision ::  aa, bb, rnorm, tol, ddot, dnrm2, temp
      external  daxpy, dcopy, ddot, dnrm2, drot, drotg, dscal

! Tolerance for size of preconditioned residual:

      pdr = 1.0d0/epsilon(pdr)

      tol   = tolrnc
      resets = 0
      first = .true.

!   alias workspace columns. 

      r  = 0
      s  = r + 1
      w  = s + 1
      y  = w
      av = y
      v  = av + 1

!   store the Givens parameters in matrix h.

      cs = restrt + 1
      sn = cs + 1

!   set initial residual (av is temporary workspace here).

      call bmux (n,n, x, work(1,av), A, colind, rwptr)
      do k = 1, n
         work(k,av) = b(k) - work(k,av)
      end do

! Return if initial guess works ok.

      urnorm = dnrm2(n,work(1,av),1)
      if (urnorm  <=  tol) go to 999
!     -------------------------------------------------
!     Adjust tolrnc if relative residual to be used:
!     -------------------------------------------------
      if (reltv) then
          tol = tol*urnorm
          tolrnc = tolrnc*urnorm
      end if

! Initial preconditioned residual

      call lusl (n, work(1,av), res, lu, jlu, uptr, pcmeth, A,  &
                 colind,rwptr) 

! Check to see if preconditioned residual is too large relative
! to unpreconditioned residual:

      rnorm = dnrm2( n, res, 1 )
      if (rnorm  >  pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

      iter = 0

   10 continue

         i = 0
         call instr(A,colind,rwptr,n, x,b, wkins, lu,jlu,uptr,pcmeth,   &
                     first,iter, iunit,instlvl)

!  construct the first column of v.

         if (iter  >  0) rnorm = dnrm2( n, res, 1 )
         temp = 1.0d0/rnorm
         do k = 1, n
            work(k,v) = temp * res(k)
         end do

!  initialize s to the elementary vector e1 scaled by rnorm.

         work( 1,s ) = rnorm
         do 20 k = 2, n
            work( k,s ) = 0.0d0
   20    continue

   30    continue

            i = i + 1
            iter = iter + 1

            call bmux (n,n, work(1,v+i-1), work(1,av), A, colind, rwptr)
            call lusl (n, work(1,av), work(1,w), lu, jlu, uptr, pcmeth,     &
                       A,colind,rwptr) 

!   construct i-th column of h orthnormal to the previous 
!   i-1 columns.

            call basis( i, n, h( 1,i ), work( 1,v ), ldw, work( 1,w ) )

!   apply Givens rotations to the i-th column of h. this
!   "updating" of the qr factorization effectively reduces
!   the hessenberg matrix to upper triangular form during
!   the restrt iterations.

            do 40 k = 1, i-1
               call drot( 1, h( k,i ), ldh, h( k+1,i ), ldh,    &
                  h( k,cs ), h( k,sn ) )
   40       continue

!   construct the i-th rotation matrix, and apply it to h so that
!   h(i+1,i) = 0.

            aa = h( i,i )
            bb = h( i+1,i )
            call drotg( aa, bb, h( i,cs ), h( i,sn ) )
            call drot( 1, h( i,i ), ldh, h( i+1,i ), ldh,   &
                  h( i,cs ), h( i,sn ) )

!   apply the i-th rotation matrix to [ s(i), s(i+1) ]'. this
!   gives an approximation of the residual norm. if less than
!   tolrnc, update the approximation vector x and quit.

            call drot( 1, work( i,s ), ldw, work( i+1,s ), ldw,     &
               h( i,cs ), h( i,sn ) )
            rnorm = abs( work( i+1,s ) )
            if ( rnorm  <=  tol ) then
               call update( i, n, x, h, ldh, work( 1,y ), work( 1,s ),      &
                  work( 1,v ), ldw )
               go to 990
            endif
            if ( iter  >=  maxits ) go to 50
            if ( i  <  restrt )   go to 30

   50    continue

!  compute current solution vector x.

!         call update( restrt, n, x, h, ldh, work( 1,y ), work( 1,s ),
!     $                work( 1,v ), ldw )
         call update( i, n, x, h, ldh, work( 1,y ), work( 1,s ),    &
                      work( 1,v ), ldw )

!  compute residual vector r, find norm, then check for tolerance.
!  (av is temporary workspace here.)

         call bmux (n,n, x, work(1,av), a, colind, rwptr)
         do k = 1, n
            work(k,av) = b(k) - work(k,av)
         end do 
         call lusl (n, work(1,av), res, lu, jlu, uptr, pcmeth,      &
                    a,colind,rwptr) 
         work( i+1,s ) = dnrm2( n, res , 1 )
         rnorm = work( i+1,s )
         if ( rnorm  <=  tol  ) go to 990
  60     continue
         if ( iter  >=  maxits ) go to 991

!  restart.

         go to 10
!
!-----------------------------------------------------------------------
!  Stopping conditions.
!-----------------------------------------------------------------------
!


! Possible convergence; check true residual 

 990   continue
       ierr = 0
       resets = resets + 1
       if (resets  >  8) then
         ierr = 2
         go to 995
       end if
       call bmux(n,n,x,work(1,v),A,colind,rwptr)
       do k = 1, n
          work(k,v) = work(k,v) - b(k)
       end do
       temp = dnrm2(n,work(1,v),1)
       if (temp  >  tolrnc) then
          tol = tol*tolrnc/temp
!          write(6,*)'>>>>> tolrnc reset to ', tol,'  <<<<<<'
          go to 60
       end if
 995   continue
       maxits = iter
      call instr(A,colind,rwptr,n, x,b, wkins, lu,jlu,uptr,     &
                     pcmeth,first,iter, iunit,instlvl)
       return

!  No convergence within allowed iterations 

 991   continue
       ierr = 3
       maxits = iter
       call instr(A,colind,rwptr,n, x,b, wkins, lu,jlu,uptr,        &
                     pcmeth,first,iter, iunit,instlvl)
       return

!  Almost exact guess on entry to routine 

 999   continue
       ierr = 1
       maxits = iter
       return
           end subroutine gmres1
!
!   ===============================================================
      subroutine update( i, n, x, h, ldh, y, s, v, ldv )

      implicit none
      integer            n, i, ldh, ldv
      double precision   x( * ), y( * ), s( * ), h( ldh,* ), v( ldv,* )
      external           daxpy, dcopy, dtrsv

!   this routine updates the gmres iterated solution approximation.      

      double precision   one
      parameter        ( one = 1.0d+0 )

!   .. executable statements ..
!
!   solve h*y = s for upper triangualar h.

      call dcopy( i, s, 1, y, 1 )
      call dtrsv( 'upper', 'notrans', 'nonunit', i, h, ldh, y, 1 )

!   compute current solution vector x = x + v*y.

      call dgemv('notrans', n, i, one, v, ldv, y, 1, one, x, 1 )

      return

      end subroutine update

!   =========================================================
      subroutine basis( i, n, h, v, ldv, w )

      implicit none
      integer            i, n, ldv
      double precision   h( * ), w( * ), v( ldv,* )

!   construct the i-th column of the upper hessenberg matrix h
!   using the gram-schmidt process on v and w.

      integer            k
      double precision   ddot, dnrm2, one
      parameter        ( one = 1.0d+0 )
      external           daxpy, dcopy, ddot, dnrm2, dscal

      do 10 k = 1, i
         h( k ) = ddot( n, w, 1, v( 1,k ), 1 )
         call daxpy( n, -h( k ), v( 1,k ), 1, w, 1 )
   10 continue
      h( i+1 ) = dnrm2( n, w, 1 )
      call dcopy( n, w, 1, v( 1,i+1 ), 1 )
      call dscal( n, one / h( i+1 ), v( 1,i+1 ), 1 )

      return

      end subroutine basis
