!=========================================================================
      subroutine qmrtf(A,colind,rwptr,n,                    &
                       x,r,b,                               &
                       tolrnc,reltv,maxits,ierr,            &
                       work,wksz,                           &
                       lu,jlu,uptr,pcmeth,iunit,instlvl)

!-----------------------------------------------------------------------
!                     *** Transpose Free QMR ***            
!                                   
!   Transpose-free QMR method of Freund and Nachtigal.  This is an  
!   implementation for small scale testing, and should not be used  
!   as a measure of the quality of all QMR methods.  In particular, 
!   users interested in QMR methods should obtain QMRPACK from netlib   
!   and experiment with the "look-ahead Lanczos" versions before    
!   rejecting QMR as a solution strategy.
!                                   
!-----------------------------------------------------------------------
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision :: array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array of additional storage.  QMR
!            requires 10 vectors of length n, and if instrumentation
!            routine instr.f is called, an additional 2 vectors of
!            length n are required.  If not enough storage is given,
!            nothing is done and an error code is returned specifying
!            the additional storage needed (in double precision words).
!  wksz    : Length of array work, in double precision words.  
!
!     (Preconditioner data structure)
!  lu,jlu  : Preconditioner in Modified Sparse Row format containing
!            the L and U factors together. The diagonal of U (stored in
!            lu(1:n) ) is inverted.  The i-th row of the lu,jlu matrix 
!            contains the i-th row of L (excluding the diagonal entry) 
!            followed by the i-th row of U.  The diagonal of L has
!            all ones, and so is not stored explicitly.
!  uptr    : Integer array of length n containing the pointers to        
!            the beginning of each row of U in the matrix lu,jlu. 
!  pcmeth  : Preconditioning method to use.  See splib.f documentation
!            for the interpretation of this integer.
!
!     (Iteration control)
!  tolrnc  : Tolerances on norm(s) used in stopping test.  This is
!            passed in as a vector in case future versions
!            set pdr and SMALL. Since Krylov methods generally
!            use the dotproduct r'*r, a local variable tolsq is used
!            in order to avoid square roots.  Note: because the
!            preconditioned residual norm || inv(M)*r || is the natural
!            one tracked by the iterative method, but users generally
!            want the true residual norm || r || as the stopping test,
!            this routine will exit the iterations when the precond
!            residual is small enough and then test the true residual.
!            If it is not sufficiently reduced, then the local tolerance
!            tolsq is reduced and iterations resume.  This is done
!            a maximum of eight times before failure is declared and
!            the routine returns.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tolrnc(1).  If it
!            is true, the test is || r_k || < || r_0 || * tolrnc(1).
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by qmrtf, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  early   : Logical parameter that indicates if a tentative exit
!            was made because of early termination.  This is needed
!            because of the way SPLIB uses unpreconditioned residuals
!            as the stopping test.
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!  SMALL   : Parameter that determines when it is not safe to divide
!            by a scalar quantity.  Error condition is set if a divisor
!            is smaller than SMALL.
!  pdr     : Parameter to determine if preconditioner is no good.
!            Experience shows that if the preconditioned residual norm
!            is more than 1/eps, where eps is the machine epsilon, then
!            even if the preconditioned residual converges, the original
!            system residual does not.  If you are not comfortable with
!            this try setting pdr = 1.0d200.  
!  urnorm  : Unpreconditioned initial residual.  Used for tests to see
!            if initial residual is good enough, and if preconditioner
!            is no good (see pdr above).
!
!  h,w,y,
!  rh,d,t,
!  v,y2,t2,
!  w2      : Integer pointers in the array work(*) that give the
!            starting position of the corresponding vectors.
!  wins1,
!  wins2   : Integer pointers in the array work(*) that give starting
!            positions of two vectors used by instrumentation routines.
!
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> convergence not achieved in maxits iterations.  
!           ierr = 4  --> Preconditioned residual is 1/eps times
!                         larger than unpreconditioned residual.
!           ierr = 5  --> convergence not achieved due to rho = 0
!           ierr = 6  --> convergence not achieved due to tau = 0
!           ierr = 7  --> convergence not achieved due to sigma = 0
!
!  maxits : Contains the number of actual iterations performed.
!  tolrnc : May be modified if reltv = .true.
!                                                                       
!======================================================================
!
!  Routines
!  --------
!
!  lusl    : Solves M z = r.
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: ddot, dcopy
!
!======================================================================
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Sun Jul 31 11:54:53 EST 1994
!
!----------------------------------------------------------------------

       implicit none

!     -------------------------
!     Declaration of arguments:
!     -------------------------
      integer n
      double precision :: a(*)
      double precision :: x(*), b(*), r(*)
      double precision :: work(n,*)
      double precision :: tolrnc(*)
      double precision :: lu(*)

      integer colind(*), rwptr(*)
      integer wksz
      integer maxits, ierr
      integer jlu(*), uptr(*)
      integer pcmeth, iunit(*), instlvl
      logical reltv

!     ------------------
!     External functions
!     ------------------
      double precision, external :: ddot,dnrm2

!     ------------- 
!     Local storage 
!     ------------- 
      logical first
      integer i, its, resets
      double precision :: SMALL, pdr, urnorm
      double precision :: alfa, beta, rho, rhonew, tau, res, tolsq
      double precision :: theta, eta, sigma, omega, thetanew, etanew, cm

!     ------------------------------
!     Pointers to work array vectors
!     ------------------------------
      integer h, w, y, rh, d, t, v, y2, t2, w2
      integer wins1, wins2

!=======================================================================
!     Start of executable statements
!=======================================================================

!     ------------------------------
!     Set up square root of safe min
!     ------------------------------
      SMALL = 1.0d0/sqrt(huge(SMALL))

!     ---------------------------------
!     Set up inverse of machine epsilon
!     ---------------------------------
      pdr = 1.0d0/epsilon(SMALL)

!     ----------------------------------------
!     Set up meaningful names for work vectors
!     ----------------------------------------
      if (instlvl .gt. 0) then
         wins1 = 1
         wins2 = wins1 + 1
      else
         wins1 = 0
         wins2 = 0
      end if
       h = wins2 + 1
       w = h + 1
       y = w + 1
      rh = y + 1
       d = rh + 1
       t = d + 1
       v = t + 1
      y2 = v + 1
      t2 = y2 + 1
      w2 = t2 + 1

!----------------------------------------------------------------
!
!     Key to work array columns if no instrumentation:
!     h    ---> work(:,1)
!     work ---> work(:,2)
!     y    ---> work(:,3)
!     rh   ---> work(:,4)
!     d    ---> work(:,5)
!     t    ---> work(:,6)
!     v    ---> work(:,7)
!     y2   ---> work(:,8)  
!     t2   ---> work(:,9)
!     work2 --> work(:,10)
!
!----------------------------------------------------------------


!     -------------------------------------
!     Stop if not enough workspace provided
!     -------------------------------------
      if (wksz .lt. (wins2+10)*n) then
         ierr = wksz - (wins2+10)*n
         maxits = 0
         return
      end if


!     -------------------------
!     Initialize for iterations
!     -------------------------
      first = .true.
      tolsq = tolrnc(1)*tolrnc(1)

!     ----------
!     r = b - Ax
!     ----------
      call bmux(n,n,x,r,A,colind,rwptr)
      do 10 i = 1,n
         work(i,h) = b(i) - r(i)
10    continue

!     -------------
!     h = inv(LU)*r
!     -------------
      call lusl(n, work(1,h), r, lu, jlu, uptr, pcmeth,a,colind,rwptr)

!     -------------------------------------------------------------
!     Check to see if preconditioned residual is too large relative
!     to unpreconditioned residual:
!     -------------------------------------------------------------
      urnorm = ddot(n, work(1,h), 1, work(1,h), 1)
      if (ddot(n,r,1,r,1) .gt. pdr*pdr*urnorm) then
         ierr = 4
         maxits = 0
         return
      end if 

!     -------------------------------------------------
!     Adjust tolerance if relative residual to be used:
!     -------------------------------------------------
      if (reltv) then
          tolsq = tolsq*urnorm
          tolrnc(1) = tolrnc(1)*sqrt(urnorm)
      end if

!     ---------------------
!     w = y = rh = r; d = 0
!     ---------------------
      do i = 1, n
         work(i,w) = r(i)
         work(i,y) = r(i)
         work(i,rh) = r(i)
         work(i,d) = 0.0d0
      end do

      call bmux(n,n,work(1,y),work(1,t),A,colind,rwptr)
      call lusl(n, work(1,t), work(1,v), lu, jlu, uptr, pcmeth,a,       &
                colind,rwptr)
      call dcopy(n, work(1,v),1,work(1,t),1)

      rho = ddot(n,work(1,rh),1,r,1)
      tau = ddot(n,r,1,r,1)

      res = tau
      tau = dsqrt(tau)
      theta = 0.0d0
      eta = 0.0d0

      resets = 0
      its = 0
      if (res.lt.tolsq) go to 900

!=======================================================================
!     Begin main loop of iterations.  Loop is equivalent to
!      do while (rnorm .gt. tolsq) .and. (its .le. maxits)
!=======================================================================

  2      continue
!
         call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,      &
                     pcmeth,first,its,iunit,instlvl)
         its = its + 1

         sigma = ddot(n,work(1,rh),1,work(1,v),1)
        if (abs(sigma) .le. SMALL) go to 950

         alfa = rho/sigma
         do i = 1,n
            work(i,y2) = work(i,y) - alfa*work(i,v)
         end do

         call bmux(n,n,work(1,y2),work(1,h),A,colind,rwptr)
         call lusl(n, work(1,h), work(1,t2), lu, jlu, uptr, pcmeth,     &
                   a,colind,rwptr)

         do i = 1,n
            work(i,w2) = work(i,w) - alfa*work(i,t)
         end do

         omega = dnrm2(n,work(1,w2),1)
        if (abs(tau) .le. SMALL) go to 940

         thetanew = omega/tau
         cm = 1/dsqrt(1+thetanew*thetanew)
         tau = omega*cm
         etanew = cm*cm*alfa
         do i = 1,n
            work(i,d) = work(i,y)+(theta*theta*eta/alfa)*work(i,d)
            x(i) = x(i) + etanew*work(i,d)
         end do
         do i = 1,n
            work(i,w) = work(i,w2)-alfa*work(i,t2)
         end do
         omega = dnrm2(n,work(1,w),1)

        if (abs(tau) .le. SMALL) go to 940

         theta = omega/tau
         cm = 1.0d0/dsqrt(1+theta*theta)
         tau = omega*cm
         eta = cm*cm*alfa
         do i = 1,n
            work(i,d) = work(i,y2) +                            &
                   (thetanew*thetanew*etanew/alfa)*work(i,d)
            x(i) = x(i)+eta*work(i,d)
         end do

         call bmux(n,n,x,r,A,colind,rwptr)
         do i = 1,n
            work(i,h) = b(i) - r(i)
         end do
         call lusl(n, work(1,h), r, lu, jlu, uptr, pcmeth,a, colind,rwptr)

         rhonew = ddot(n,work(1,rh),1,work(1,w),1)
         res = ddot(n,r,1,r,1)

         if (abs(rho) .le. SMALL) go to 930

         beta = rhonew/rho
         rho = rhonew
         do i = 1,n
            work(i,y) = work(i,w)+beta*work(i,y2)
         end do
         call bmux(n,n,work(1,y),work(1,h),A,colind,rwptr)
         call lusl(n, work(1,h), work(1,t), lu, jlu, uptr, pcmeth,      &
                   a,colind,rwptr)
         do i = 1,n
            work(i,v) = work(i,t) + beta*(work(i,t2)+beta*work(i,v))
         end do

         if (res.lt.tolsq) go to 910
  60     continue
         if (its .ge. maxits) go to 920

      go to 2

!=======================================================================
!           Stopping conditions
!=======================================================================

!     =====================
!     (Almost) exact guess
!     =====================
 900  continue
      ierr = 1
      maxits = 0
      return

!     =========================================
!     Possible convergence, check true residual
!     =========================================
 910  continue
      ierr = 0
      resets = resets + 1
      if (resets .gt. 8) then
         ierr = 2
         go to 915
      end if
      call bmux(n,n,x,work(1,h),A,colind,rwptr)
      do i = 1, n
         work(i,h) = b(i) - work(i,h)
      end do
      res = ddot(n,work(1,h),1,work(1,h),1)
      if (res .gt. tolrnc(1)*tolrnc(1)) then
          tolsq = tolsq*tolrnc(1)*tolrnc(1)/res
!          write(6,*)'>>>>> tolerance reset to ', sqrt(tolsq),'  <<<<<<'
          go to 60
      end if
 915  continue
      maxits = its
      call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its,iunit,instlvl)
      return

!     =========================================
!     No convergence in max allowed iterations
!     =========================================
 920  continue
      ierr = 3
      maxits = its
      call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its,iunit,instlvl)
      return

!     =======
!     rho = 0
!     =======
 930  continue
      ierr = 5
      maxits = its
      call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its,iunit,instlvl)
      return

!     =======
!     tau = 0
!     =======
 940  continue
      ierr = 6
      maxits = its
      call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its,iunit,instlvl)
      return

!     =========
!     sigma = 0
!     =========
 950  continue
      ierr = 7
      maxits = its
      call instr(a,colind,rwptr,n, x,b, work(1,wins1), lu,jlu,uptr,     &
                     pcmeth,first,its,iunit,instlvl)
      return

      end subroutine qmrtf
