!======================================================================== 
        subroutine sor(A,colind,rwptr,n,                        &
                       x,r,b,                                   &
                       tol,reltv,maxits,ierr,                   &
                       work,wksz,                               &
                       omega,jlu,diag,pcmeth,iunit,instlvl)


!======================================================================== 
!                *** SOR Iteration ***                  
!                ----------------------                 
!                                   
!  Successive overrelaxation. This is not meant to be a serious     
!  implementation, and is provided because it is useful in teaching 
!  courses in iterative methods.  There are severed limitations:    
!     1) The matrix in CSR format must have its rows ordered so that    
!        in each row first all of the lower triangular, then the    
!        diagonal, and finally the upper triangular entries occur.  
!     2) No preconditioning is allowed with this implementation.    
!        not a limitation of the alogrithm, but of the current SPLIB.   
!     3) Tolerances (tolrnc(*)) is replaced by scalar tol.          
!     4) The data structure for A may be changed, to have conditions    
!        listed in (1) above satisfied.  Done by calling ssor.f     
!                                   
!  This routine also differs from the usual calling sequence of SPLIB.  
!  Since no preconditioning is allowed, what is formerly the array  
!  alu is replaced here with the relaxation parameter omega.        
!  The array lu is used here to set up pointers to the          
!  diagonal entries in each row, so must be of length n.  This makes    
!  required upper/lower triangular solves more efficienct.      
!                                   
!  The stopping test in the iterations is with the change to x, since   
!  this can be estimated within the iterations cheaply.         
!  This is checked against the 2-norm when the condition is satisfied.  
!                                   
!=======================================================================
!
!  Arguments:
!  ----------
!
!     (Data structure for A)
!  n       : Order of matrix A.
!  A       : double precision array for nonzero entries of matrix A; of
!            length nnz = rwptr(n+1)-1.
!  colind  : Integer array of column indices for entries in A; of
!            length nnz = rwptr(n+1)-1.
!  rwptr   : Integer array of pointers to beginning position of rows
!            in arrays A and colind.  Must be of length n+1.
!
!     (Required vectors)
!  x       : Vector of unknowns to be solved for.  Contains initial
!            estimate of x on entry, and contains final approximate
!            on return.
!  r       : Residual vector r = b - A*x.
!  b       : Right hand side vector of system to be solved: Ax = b
!  work    : Double precision array; used here only if instrumentation
!            is turned on.
!  wksz    : Length of array work, in double precision words.  
!            Must be >= 2n if instrumentation is used.
!
!     (Preconditioner data structure)
!  omega   : Relaxation parameter.
!  jlu     : Dummy argument here.
!  diag    : Integer array of length n used to contain pointers to
!            diagonal entries in A's data structure.
!  pcmeth  : Preconditioning method to use.  WARNING: this version of
!            SOR does not allow preconditioning.
!
!     (Iteration control)
!  tol     : Tolerance on norm used in stopping test.
!  reltv   : Logical variable indicating whether or not to use a
!            relative residual as the stopping test.  If this is
!            false, the stopping test is || r_k || < tol.  If it
!            is true, the test is || r_k || < || r_0 || * tol.
!            Note that it is possible to modify SPLIB to use other
!            stopping tests (see routine instr.f).
!  maxits  : Maximum number of iterations allowed.
!
!      (Other stuff)
!  ierr    : Error flag; see below.
!  iunit   : Integer vector of unit numbers for the instrumentation
!            routine instr.f to use.  Not referenced by cgne, just
!            passed along to instr.f
!  instlvl : Integer giving instrumentation level.  If instlvl = 0,
!            then instr.f is not called and array work() need not
!            provide the 2n double precision words used by instr.f
!
!  Some internal variables:
!  ------------------------
!
!  first   : Logical parameter that specifies to instr.f if this is
!            the first iteration, and so if new headers need to be
!            written to the Xgraph output files.
!======================================================================
!
! Return values:                                                           
! --------------
!
!  x      : Contains an approximate solution (upon successful return).  
!
!  ierr   : Error flag with the following meanings:          
!           ierr = -m --> Not enough workspace provided; m more double
!                         precision words needed in array work().
!           ierr = 0  --> Successful return; tolerance achieved.
!           ierr = 1  --> The initial guess satisfies the stopping test.
!                         This is often a clue that rhs not set properly.
!           ierr = 2  --> Tolerance reset 8 times in attempt to
!                         get true residual small enough.
!           ierr = 3  --> Convergence not achieved in maxits iterations.  
!           ierr = 4  --> Diagonal entry of A not safely invertible.
!
!  maxits : Contains the number of actual iterations performed.
!  tol    : May be modified if reltv = .true.
!
!  diag   : Integer vector of pointers to diagonal entries in 
!           the data structure for A.
!
!  A      : Although the *matrix* A is not changed by this routine,
!           it may reorder the matrix to have each row in column 
!           increasing order.  So A, colind may both be changed.  This
!           is primarily a problem if you have computed a preconditioner
!           when using another method, and want to follow a call to
!           gauss.f with a call to another solver that uses the
!           previously computed preconditioner.  BEWARE!
!                                                                       
!======================================================================
!  Routines
!  --------
!
!  bmux    : Computes matrix-vector product with A.
!  instr   : Instrumentation routine for timing, etc.
!  BLAS routines: dnrm2
!
!=======================================================================
!                                   
!     Randall Bramley                           
!     Department of Computer Science                    
!     Indiana University                        
!     email: bramley@cs.indiana.edu                 
!     Wed Sep 21 07:53:21 EST 1994                  
!     Documented: Mon Jul 10 10:56:01 EST 1995              
!                                   
!=======================================================================
!
        implicit none

!       -------------------------
!       Declaration of arguments:
!       -------------------------
        double precision :: omega
        integer :: wksz, jlu, pcmeth
        logical :: first, reltv
        integer :: iunit(*), instlvl

        integer :: n, rwptr(n+1), colind(*), diag(n), ierr
        double precision :: work(n,*), A(*), tol
        double precision :: x(n), b(n), r(n)
        integer :: maxits

!       ---------------
!       Local variables
!       ---------------
        double precision :: rho, toloc, t, s
        integer :: i,k,iters, resets

!       ------------------
!       External functions
!       ------------------
        double precision, external :: dnrm2

!=======================================================================
!     Start of executable statements
!=======================================================================

        ierr = 0

!       --------------------------- 
!       Check for adequate worksize 
!       --------------------------- 
        if (instlvl > 0 .and. wksz < 2*n) then
               ierr = wksz - 2*n 
               return
        end if

!       ---------------------------------------------------------
!       First, massage the data structure of A to get pointers to
!       diagonal entries.  Use the ssor.f routine for this:
!       ---------------------------------------------------------
        call ssor (n,A,colind,rwptr,omega,omega,diag,wksz,ierr)

!       -----------------------------------------------
!       Return if some diagonal entry of A is too small
!       -----------------------------------------------
        if (ierr <  0) then
           ierr = 4
           return
        end if

!       ------------------------------------
!       Now matrix data structure is set up.
!       Compute the initial residual norm.  
!       ------------------------------------
        call bmux(n, n, x, r, A, colind, rwptr)
        do i = 1, n
           r(i) = b(i) - r(i)
        end do
        rho = dnrm2(n,r,1)
        if (reltv) toloc = tol*rho

!       --------------------------
!       Initialize for iterations.
!       --------------------------
        first = .true.
        iters = 0
        resets = 0

!       =================
!       Outer loop begins
!       =================

 80     continue

        rho = 0.0d0
        call instr(A,colind,rwptr,n, x,b, work, omega,jlu,diag,pcmeth,  &
                     first,iters,iunit,instlvl)
        do k = 1, n
           t = 0.0d0
           do i = rwptr(k), diag(k)-1
              t = t + A(i)*x(colind(i))
           end do
           do i = diag(k)+1, rwptr(k+1)-1
              t = t + A(i)*x(colind(i))
           end do
           s  = (b(k) - t)/A(diag(k))
           rho  = rho + abs( x(k) - s )
           x(k) = x(k) + omega * (s - x(k))
        end do 

        iters = iters + 1

        if (rho .gt. toloc .and. iters .lt. maxits) go to 80

        if (iters .ge. maxits) then
           ierr = 3
           return
        end if 

        if (rho .le. toloc) then
           call bmux(n, n, x, r, A, colind, rwptr)
           do i = 1, n
              r(i) = b(i) - r(i)
           end do
           t = dnrm2(n,r,1)
           if (t .le. tol) then
              maxits = iters
              return
           end if 
           resets = resets + 1
           if (resets .gt. 8) then
              ierr = 2
              return
           end if
           toloc =  0.9 * toloc * tol/t
           go to 80
        end if 

        return

        end subroutine sor
