!========================================================================        
!                    **** utilio.f  ****
!========================================================================        
!
!  utilio.f contains routines and functions for I/O and related stuff.
!  Currently contains the routines:
!      unpack  : reads in and unpacks control parameters
!      strings : sets up character strings that name the method 
!                and preconditioner used; for Xgraph output files
!
!      matchk  : checks matrix and inserts diagonal entries if needed.
!      readf   : read in a vector stored in ascii, one entry per row
!      readmt3 : version of readmt from SPARSKIT which uses pointers into
!                a work array, handles rectangular matrices, and provides
!                space for diagonal entry in each row if specified.
!      rdline  : reads in a single line of input for readmt3
!      readunf2: Read a H/B matrix stored in an unformatted file.
!      prtval  : Print out a vector (int or dbl)
!      prtmt   : Print out a Harwell/Boeing matrix (this is from Sparskit)
!      descr   : Writes out an HTML file describing SPLIB
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     Phone: (812) 855-7790
!     email: bramley@cs.indiana.edu
!     Mon Dec 16 11:33:48 EST 1996
!     Updated to include descr, Fri Apr 16 14:30:25 EST 1999
!
!
!========================================================================        

      subroutine strings(solmeth, solver, pcmeth, precond)
!-----------------------------------------------------------------------
!  Set up character strings identifying the solver/precond pair as
!  needed in SPLIB's Xgraph output files.
!-----------------------------------------------------------------------
      implicit none
      character solver*5, precond*4
      integer :: solmeth, pcmeth

      if( solmeth  ==  1) then
         solver = 'BiCG '
      elseif (solmeth  ==  2) then
         solver = 'CGNE '
      elseif (solmeth  ==  3) then
         solver = 'CGNR '
      elseif (solmeth  ==  4) then
         solver = 'CG-Sq'
      elseif (solmeth  ==  5) then
         solver = 'CGStb'
      elseif (solmeth  ==  6) then
         solver = 'GMRES'
      elseif (solmeth  ==  7) then
         solver = 'QMRTF'
      elseif (solmeth  ==  8) then
         solver = 'BiCGS'
      elseif (solmeth  ==  9) then
         solver = 'GMRST'
      elseif (solmeth  ==  10) then
         solver = 'JACOB'
      elseif (solmeth  ==  11) then
         solver = 'GAUSS'
      elseif (solmeth  ==  12) then
         solver = 'SOR'
      elseif (solmeth  ==  13) then
         solver = 'ORTM'
      elseif (solmeth  ==  14) then
         solver = 'CSCGS'
      end if

      if (pcmeth  ==  1) then
        precond = 'ILU '
      elseif (pcmeth  ==  2) then
        precond = 'MILU'
      elseif (pcmeth  ==  3) then
        precond = 'ILUT'
      elseif (pcmeth  ==  4) then
        precond = 'SSOR'
      elseif (pcmeth  ==  5) then
        precond = 'TRID'
      elseif (pcmeth  ==  6) then
        precond = 'ILU0'
      elseif (pcmeth  ==  7) then
        precond = 'EMGS'
      elseif (pcmeth  ==  0) then
        precond = 'NOPC'
      end if

      return
      end
!========================================================================

       subroutine matchk(A, ja, ia, n, diag, ndiag)
!------------------------------------------------------------------------
!
!  Check the matrix A in CSR format to make sure it includes all diagonal
!  entries.  If it does not have them, insert 0.0 into the data structure
!  for each missing diagonal entry.  At the same time, sort the rows of
!  A to have entries in increasing column order.  This is useful for 
!  preconditioning methods like SSOR or ILU(-1), and for matrix splitting
!  based iterative methods like Jacobi.
!
!  a, ja, ia    : matrix A in CSR format
!  n        : order of matrix A
!  diag     : integer work vector of length n.  On return, diag(1:n)
!                 contains pointers to the index of each diagonal entry
!                 in the modified CSR data structure.
!  ndiag    : integer, on return gives number of diag entries that
!                 were missing from original matrix data structure.
!  
!------------------------------------------------------------------------
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Thu Aug 25 08:06:44 EST 1994
!
!------------------------------------------------------------------------
       use kinds
       implicit none

!      ------------------------
!      Declaration of arguments
!      ------------------------
       real(real8)  :: A(*)
       integer :: n, ja(*), ia(n+1), diag(n), ndiag

!      ---------------
!      Local variables
!      ---------------
       integer :: row, lastptr, ibegin, iend, j, lp, ap, added

!      ----------------------------------------------
!      Value to insert for missing diagonal entries
!      Must be nonzero for ancient stationary methods
!      ----------------------------------------------
       real(real8), parameter :: INSERT = 0.0_real8

!=======================================================================
!          Beginning of Executable Statements
!=======================================================================


!      ----------------------------------------------------------------
!      Find diagonal entry for each row.  If it exists, set diag(row)
!      to be its index in CSR data structure.  Otherwise, set diag(row)
!      to be zero.
!      ----------------------------------------------------------------
       ndiag = 0
       do row = 1, n
          diag(row) = 0
          do j = ia(row), ia(row+1)-1
             if (ja(j)  ==  row) then
                diag(row) = j
!               A(j) = max(A(j),1.0d0)
                go to 20
             end if 
          end do
          ndiag = ndiag + 1
 20    continue
       end do

!      --------------------------------------------------------
!      If ndiag = 0 here, then all diagonal entries are present
!      in the data structure.  Sort the rows and exit.
!      --------------------------------------------------------
       if (ndiag  ==  0) then
          do row = 1, n
             ibegin = ia(row)
             iend   = ia(row+1)
             call rowinc( iend - ibegin, ja(ibegin), A(ibegin) )
          end do
          return
       end if

!      ---------------------------------------------------------
!      Otherwise, insert some diagonal entries into data struct.
!      ---------------------------------------------------------
       lastptr = ia(n+1) - 1 + ndiag
       lp = lastptr
       ap = ia(n+1) - 1

       do row = n, 1, -1

          ibegin = ia(row)
          iend   = ia(row+1)

!         --------------------------------------------
!         Copy a, ja over, shifting to right as needed
!         --------------------------------------------
          do j = iend-1, ibegin, -1
             A(lp)  = A(ap)
             ja(lp) = ja(ap)
             lp = lp - 1 
             ap = ap - 1 
          end do 

!         --------------------------------
!         Insert diagonal entry if missing
!         --------------------------------
          added = 0
          if (diag(row)  ==  0) then
             A(lp) = INSERT
             ja(lp) = row
             lp  = lp - 1
             added = 1
          end if 

!         ----------------------------
!         Sort newly modified row of A
!         ----------------------------
          call rowinc( iend - ibegin + added, ja(lp+1), A(lp+1) )

!         -------------------
!         Set new row pointer
!         -------------------
          ia(row+1) = lastptr + 1

!         -------------------
!         Change last pointer
!         -------------------
          lastptr = lp

       end do
       return
       end

!
!========================================================================        
!
      subroutine readf(iounit,n,rarray)
!-----------------------------------------------------------------------
!  Read in a right hand side vector f.
!-----------------------------------------------------------------------
      integer :: iounit, n
      double precision :: rarray(n)
      read(iounit,*) (rarray(i), i = 1, n)
      return
      end subroutine readf
!
!=======================================================================
!
      subroutine readmt3(job,iounit,A,colind,rwptr,b,sol,g,nrhs, &
               guesol,nrow,ncol,nnz,title,key,rua,ierr, chkdiag)
!-----------------------------------------------------------------------
!
! This routine reads in the transpose of a H/B file, and is a modification
! of the original SPARSKIT routine readmt.
! readmt3 also assures that the matrix data structure includes
! space for the diagonal entries by adjusting it as we go along, if the
! argument chkdiag = .true.
!
! Note that although this is used in SPLIB to read matrices stored in
! CSR format, it works equally well on standard H/B format matrices
! (which are in CSC format).
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University, Bloomington
!     Mon Aug  8 12:25:22 EST 1994
!     Revised to do F95 allocation of arrays,
!       Mon Jun 21 14:05:42 EST 2004
!
!-----------------------------------------------------------------------
!
!   Control Arguments:
!   ==================
!
! job    = integer to indicate what is to be read. (note: job is an
!          input and output parameter, it can be modified on return)
!          job = 0    read the values of ncol, nrow, nnz, title, key,
!                     rua and return.  Matrix is not read and arrays
!                     A, colind, rwptr, rhs are not allocated.
!          job = 1    read structure only, i.e., the arrays colind and rwptr.
!          job = 2    read matrix including values, i.e., A, colind, rwptr
!          job = 3    read matrix and right hand sides: A,colind,rwptr,rhs.
!             rhs may contain initial guesses and exact 
!                     solutions appended to the actual right hand sides.
!             this will be indicated by the output parameter
!                     guesol [see below]. 
!                     
! iounit = logical unit number from which to read the matrix.
! work   = integer array, used for storing the vectors returned.
! next   = integer pointer, indicating next available position in
!          array work.
!
!   ============== 
!   Return Values:
!   ============== 
!
! job    = on return job may be modified to the highest job it could
!          do: if job=2 on entry but no matrix values are available it
!          is reset to job=1 on return. Similarly if job=3 but no rhs 
!          is provided then it is reset to job=2 or job=1 depending on 
!          whether or not matrix values are provided.
!          Note that no error message is triggered (i.e. ierr = 0)
!          on return in these cases.  It is therefore important to
!          compare the values of job on entry and return.
!
! A  = the A matrix in the a, rwptr, colind (column) storage format
! colind = column number of element a(i,j) in array a.
! rwptr  = pointer array. rwptr(i) points to the 
!                        beginning of column i, and rwptr(nrow+1) points to
!                        one position beyond the end of arrays a, colind.
!
! rhs    = (int pointer) real array of size nrow + 1 if available (see job)
!
! nrhs   = integer containing the number of right-hand sides found
!          each right hand side may be accompanied with an initial guess
!          and also the exact solution.
!
! guesol = a 2-character string indicating whether an initial guess 
!          (1-st character) and / or the exact solution (2-nd
!          character) is provided with the right hand side.
!      If the first character of guesol is 'G' it means that an
!          an initial guess is provided for each right-hand side.
!          These are appended to the right hand-sides in the array rhs.
!      If the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are  appended to the right hand-sides 
!          and the initial guesses (if any) in the array rhs.
!
! nrow   = number of rows in matrix
! ncol   = number of columns in matrix 
! nnz    = number of nonzero elements in A. This info is returned
!          even if there is not enough space in a, colind, rwptr, in order
!          to determine the minimum storage needed. 
!
! title  = (character*72) title of test matrix 
! key    = (character*8)  key of matrix 
! rua   = (charatcer*3)  type of matrix.
!          For meaning of title, key and type refer to documentation 
!          Harwell/Boeing matrices.
!
! ierr   = integer used for error messages 
!         ierr  =  0 --> the matrix has been read normally. 
!         ierr  =  1 --> the array matrix could not be read 
!                        because inadequate storage provided.
!         ierr  =  2 --> the rhs/guess/sol could not be read.
!         ierr  =  3 --> a sparse rhs was specified.
!
! mneed  = integer; amount of extra storage needed to get all info.
!          If positive, gives amount of storage left over after reading
!          in the matrix and rhs data.  If negative, gives the additional
!          number of integer words needed to successfully read stuff in.
! 
! Notes:
!-------
! 1) The file inout must be open (and possibly rewound if necessary)
!    prior to calling readmt.
! 2) Refer to the documentation on the Harwell-Boeing formats
!    for details on the format assumed by readmt.
!    We summarize the format here for convenience.
!  
!    a) All lines in inout are assumed to be 80 character long.
!    b) The file consists of a header followed by the block of the 
!       column start pointers followed by the block of the
!       row indices, followed by the block of the real values and
!       finally the numerical values of the right-hand-side if a 
!       right hand side is supplied. 
!    c) The file starts by a header which contains four lines if no
!       right hand side is supplied and five lines otherwise.
!       * first line contains the title (72 characters long) followed by
!         the 8-character identifier (name of the matrix, called key)
!        [ A72,A8 ]
!       * second line contains the number of lines for each
!         of the following data blocks (4 of them) and the total number 
!         of lines excluding the header.
!        [5i4]
!       * the third line contains a three character string identifying
!         the type of matrices as they are referenced in the Harwell
!         Boeing documentation [e.g., rua, rsa,..] and the number of
!         rows, columns, nonzero entries.
!         [A3,11X,4I14]
!       * The fourth line contains the variable fortran format
!         for the following data blocks.
!         [2A16,2A20] 
!       * The fifth line is only present if right-hand-sides are 
!         supplied. It consists of three one character-strings containing
!         the storage format for the right-hand-sides 
!         ('F'= full,'M'=sparse=same as matrix), an initial guess 
!         indicator ('G' for yes), an exact solution indicator 
!         ('X' for yes), followed by the number of right-hand-sides
!         and then the number of row indices. 
!         [A3,11X,2I14] 
!     d) The three following blocks follow the header as described 
!        above.
!     e) In case the right hand-side are in sparse formats then 
!        the fourth block uses the same storage format as for the matrix
!        to describe the NRHS right hand sides provided, with a column
!        being replaced by a right hand side.
!-----------------------------------------------------------------------

      implicit none

!     Declarations of variables
!     =========================

      character title*72, key*8, rua*3, ptrfmt*20, indfmt*20,      &
             valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer :: totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,   &
           nnz, neltvl, nrhs, nrwindx
      integer :: job, iounit, ierr
      integer :: length, nvec, next, remain, ndiag, row, j

!     arrays
!     -------
      double precision, pointer      :: A(:), b(:), sol(:), g(:)
      integer, pointer               :: rwptr(:), colind(:)

!     Insert diagonal entries or not?
!     -------------------------------
      logical chkdiag

!=======================================================================
!        Start of Executable Statements
!=======================================================================

!     -------------------
!     Read in header info
!     -------------------
      rewind(iounit)
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd,  &
           rhscrd, rua, nrow, ncol, nnz, neltvl, ptrfmt, indfmt,   &
           valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)

!     -----------------------------------------------
!     Read in right hand side formats if rhs provided
!     -----------------------------------------------
      if (rhscrd  >  0) read(iounit,11) rhstyp, nrhs , nrwindx
 11   format (a3,11x,i14,i14)


!     -----------------------
!     Anything else to read ?
!     -----------------------
      ierr = 0
      if (job  <=  0) return
      next = 1

!     -----------------------------------
!     Read row pointer and column numbers
!     -----------------------------------
      read(iounit,ptrfmt) (rwptr(j), j = 1, nrow+1)
      read(iounit,indfmt) (colind(j), j = 1, nnz)

!     ------------------------------------------------------
!     Check to see if any diagonal entries of A are missing,
!     if it is required to force diagonal entries to be
!     present.
!     ------------------------------------------------------
      ndiag = 0
      if (chkdiag) then
         do row = 1, nrow
            do j = rwptr(row), rwptr(row+1)-1
               if (colind(j)  ==  row) go to 20
            end do 
            ndiag = ndiag + 1
 20      continue
         end do
         nnz = ndiag + nnz
      end if 

!     ---------------------------------
!     Read values of matrix if required
!     ---------------------------------
      if (job <= 1)  return

      if (valcrd  <=  0) then
      job = 1
      return
      endif

!     -------------------------
!     Now read in matrix values
!     -------------------------
      read(iounit,valfmt) A(1:nnz-ndiag)

!     ------------------------
!     Read rhs if required ... 
!     ------------------------
      if (job  <=  2)  return
!     --------------------
!     ... and if available 
!     --------------------
      if ( rhscrd  <=  0) then
      job = 2
      return
      endif

!     ----------------------------------------------------
!     Read right-hand-side vector b, a guess if available,
!     and a true solution vector if available.
!     ----------------------------------------------------
      if (rhstyp(1:1)  ==  'M') then 
         ierr = 3
         return
      endif
      guesol = rhstyp(2:3) 
      nvec = 1
      if (guesol(1:1)  ==  'G' .or. guesol(1:1)  ==  'g')   &
                    nvec=nvec+1
      if (guesol(2:2)  ==  'X' .or. guesol(2:2)  ==  'x')   &
                    nvec=nvec+1

!     ---------------------
!     Read right-hand-sides
!     ---------------------
      length = nrhs*nrow
      read(iounit,rhsfmt) b(1:length)

!     ---------------------------------
!     Read initial guesses if available
!     ---------------------------------
      if (guesol(1:1)  ==  'G' .or. guesol(1:1)  ==  'g') then
        read(iounit,rhsfmt) g(1:length)
      endif

!     ---------------------------------
!     Read exact solutions if available
!     ---------------------------------
      if (guesol(2:2)  ==  'X' .or. guesol(2:2)  ==  'x') then
         read(iounit,rhsfmt) sol(1:length)
      endif

      return
      end subroutine readmt3

!========================================================================        

       subroutine unpack(ireltv,  rpcprm    ,pcmeth  ,          &
                         ipcprm   ,solmeth ,krylov  ,           &
                         tol     ,maxits  , iredo,              &
                         iunit   ,                              &
                         instlvl , scale, fmtd,                 &
                         title, key, solver, precond)
!========================================================================        
!
!   Read in parameters that define stopping tolerances, methods to be
!   used.  Unpack information into pcmeth, ipcprm, etc.  Stop if
!   solmeth = 0.
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     email: bramley@cs.indiana.edu
!     Wed Jan 26 13:23:46 EST 1994
!
!========================================================================        

      use kinds
      implicit none

      integer       :: pcmeth, ipcprm, solmeth
      integer       :: krylov, maxits
      integer       :: iunit(20)
      real(real8)   :: rpcprm, tol
      logical       :: results
      integer       :: ireltv, iredo
      integer       :: fmtd
      integer       :: instlvl, scale
      character     :: title*72, key*8
      character     :: solver*5, precond*4

!=======================================================================
!      Start of Executable Statements
!=======================================================================

      results = (iunit(1) .ne. 0)
      read(iunit(2),*) solmeth

!     -----------------------------------------------------------
!     Sentinel value of solmeth = 0 is used for stopping the run.
!     -----------------------------------------------------------
      if (solmeth  <=  0) stop

      read(iunit(2),*) krylov
      read(iunit(2),*) pcmeth
      read(iunit(2),*) ipcprm
      read(iunit(2),*) rpcprm

      if (pcmeth  ==  7) then
         if (rpcprm  >  1.0d0) ipcprm = int(rpcprm)
         if (rpcprm  <  1.0d0 .and. rpcprm  >  0.0d0) ipcprm = -2
         if (rpcprm  ==  0.0d0) ipcprm = 0
         if (rpcprm  ==  1.0d0) ipcprm = -1
      end if

      read(iunit(2),*) tol, ireltv
      read(iunit(2),*) maxits
      read(iunit(2),*) instlvl
      read(iunit(2),*) scale
      read(iunit(2),*) fmtd
      read(iunit(2),*) iredo

!     ------------------------------------
!     Check input parameters for validity:
!     ------------------------------------
!      if (tol  <  0.0) then
!         if (results) write(iunit(1),4040) tol
!         tol = 1.0d-6
!      end if
!      if (rpcprm  <  0.0) then
!         if (results) write(iunit(1),4045) rpcprm
!         stop
!      end if
!      if (solmeth <  0 .or. solmeth  >  14) then
!         if (results) write(iunit(1),4010) solmeth
!         stop
!      end if
!      if (pcmeth  <  0 .or. pcmeth  >  7) then
!         if (results) write(iunit(1),4020) pcmeth
!         stop
!      end if
!      if (scale  <  0 .or. scale  >  4) then
!         if (results) write(iunit(1),4050) scale
!         stop
!      end if
!
!  Set up character strings identifying the solver/precond pair:
!
      call strings(solmeth, solver, pcmeth, precond)

!       =================
!       Format Statements
!       =================
 4010  format(' *** Invalid solution method parameter: solmeth = ',i5)
 4020  format(' *** Invalid pc method parameter: pcmeth = ',i5)
 4050  format(' *** Invalid scaling method parameter: scale = ',i2)
 4040  format(' *** Invalid tolerance: tol = ',1pe12.5,/,           &
              '     tol being reset to 1.0e-6')
 4045  format(' *** Invalid parameter: s = ',1pe12.5)

    return
    end subroutine unpack

!======================================================================
      subroutine prtval(iunit,rua,n,iarray,rarray)
!----------------------------------------------------------------------
!
!  Print out n values of the array iarray or rarray, as
!  specified by the character array rua.
!
!----------------------------------------------------------------------
      use kinds
      implicit none
      integer       :: n, iarray(n), i, iunit
      real(real8)   :: rarray(n)
      character(len=3)  :: rua
      if (rua  ==  'int') then
         write(iunit,1000) (iarray(i), i = 1, n)
      end if
      if (rua  ==  'dbl') then
         write(iunit,1010) (rarray(i), i = 1, n)
      end if
 1000 format(1(i7,3x))
 1010 format(1(d24.16,3x))
      return
      end subroutine prtval

!======================================================================
      subroutine readunf2(nrow,ncol,nnz,a,ja,ia,iounit,job,  &
                        ierr,nrhs,rhs,title,key,rua,precs)
!-----------------------------------------------------------------------
! This subroutine reads a matix stored in binary CSR format.
! This version has a "job" flag, allowing sizes to be found and returned,
! allowing the calling routine to allocate memory for the arrays.
!
!  NOTE: this assumes nnz is part of the binary formatted file, unlike
!  earlier (pre-2004) versions!
!
!-----------------------------------------------------------------------
! On entry:
!  iounit: the I/O unit that opened for reading.
!  nrhs  : 0 if no right hand side vector is to be read, nonzero
!          otherwise.  nrhs is reset to zero if no rhs vector is
!          available to be read.  This is not an error condition, so
!          if you request a rhs to be read, be sure to check nrhs on
!          return to see if one was available.
!  precs: number of significant digits for representing floats.
! On return:
!    nrow: the actual number of rows in array.
!    ncol: the actual number of cols in array.
!     nnz: the actual number of non-zero entries.
!    nrhs: the actual number of right hand side vectors read (0 or 1)
! ia,ja,a: pointers in work array to where corr arrays go.
!    ierr: a error code, it's same as that used in reaadsk
!          0 -- OK
!          1 -- error in reading iounit
!          2 -- end-of-file reached while reading data file
!          3 -- not enough storage provided; check mneed
! On error:
!     mneed contains additional storage needed if not enough was provided.
!-----------------------------------------------------------------------
!
      integer, parameter:: intdbl = 2
      integer :: nrow,ncol, iounit, nnz, k, precs
      integer :: job
      integer, pointer, dimension(:) :: ia, ja
      integer :: next, mneed
      double precision, pointer, dimension(:) :: A, rhs
      character :: title*72, key*8, rua*3

      ierr = 0
      rewind iounit

      read(unit=iounit, err=1000, end=1010) nrow, ncol, nnz
      read(unit=iounit, err=1000, end=1010) title, key, rua, precs, nrhs

      if (job <= 0) return
      read(unit=iounit, err=1000, end=1010) (ia(k),k=1,nrow+1)
      read(unit=iounit, err=1000, end=1010) (ja(k),k=1,nnz)
      read(unit=iounit, err=1000, end=1010) ( A(k),k=1,nnz)
      if (nrhs  <=  0)  return
      read(unit=iounit, err=1000, end=1010) (rhs(k),k=1,nrow)
      return

 1000 ierr = 1
      return
 1010 ierr = 2
      return
      end subroutine readunf2

!======================================================================
      subroutine prtmt (nrow,ncol,a,ja,ia,rhs,guesol,title,key,rua,    &
           ifmt,job,iounit)
!-----------------------------------------------------------------------
! writes a matrix in Harwell-Boeing format into a file.
! assumes that the matrix is stored in COMPRESSED SPARSE COLUMN FORMAT.
! some limited functionality for right hand sides. 
! Author: Youcef Saad - Date: Sept., 1989 - updated Oct. 31, 1989 to
! cope with new format. 
!-----------------------------------------------------------------------
! on entry:
!---------
! nrow   = number of rows in matrix
! ncol   = number of columns in matrix 
! a  = double precision :: array containing the values of the matrix stored 
!          columnwise
! ja     = integer array of the same length as a containing the column
!          indices of the corresponding matrix elements of array a.
! ia     = integer array of containing the pointers to the beginning of 
!      the row in arrays a and ja.
! rhs    = real array  containing the right-hand-side (s) and optionally
!          the associated initial guesses and/or exact solutions
!          in this order. See also guesol for details. the vector rhs will
!          be used only if job  >  2 (see below). Only full storage for
!          the right hand sides is supported. 
!
! guesol = a 2-character string indicating whether an initial guess 
!          (1-st character) and / or the exact solution (2-nd)
!          character) is provided with the right hand side.
!      if the first character of guesol is 'G' it means that an
!          an intial guess is provided for each right-hand sides. 
!          These are assumed to be appended to the right hand-sides in 
!          the array rhs.
!      if the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are assumed to be appended to the right hand-sides 
!          and the initial guesses (if any) in the array rhs.
!
! title  = character*72 = title of matrix test ( character a*72 ).
! key    = character*8  = key of matrix 
! rua   = charatcer*3  = rua of matrix.
!
! ifmt   = integer specifying the format chosen for the real values
!      to be output (i.e., for a, and for rhs-guess-sol if 
!          applicable). The meaning of ifmt is as follows.
!     * if (ifmt  <  100) then the D descriptor is used,
!           format Dd.m, in which the length (m) of the mantissa is 
!           precisely the integer ifmt (and d = ifmt+6)
!     * if (ifmt  >  100) then prtmt will use the 
!           F- descriptor (format Fd.m) in which the length of the 
!           mantissa (m) is the integer mod(ifmt,100) and the length 
!           of the integer part is k=ifmt/100 (and d = k+m+2)
!       Thus  ifmt= 4   means  D10.4  +.xxxxD+ee    while
!             ifmt=104  means  F7.4   +x.xxxx
!             ifmt=205  means  F9.5   +xx.xxxxx
!       Note: formats for ja, and ia are internally computed.
!
! job    = integer to indicate whether matrix values and
!      a right-hand-side is available to be written
!          job = 1   write srtucture only, i.e., the arrays ja and ia.
!          job = 2   write matrix including values, i.e., a, ja, ia
!          job = 3   write matrix and one right hand side: a,ja,ia,rhs.
!      job = nrhs+2 write matrix and nrhs successive right hand sides
!      Note that there cannot be any right-hand-side if the matrix
!      has no values. Also the initial guess and exact solutions when 
!          provided are for each right hand side. For example if nrhs=2 
!          and guesol='GX' there are 6 vectors to write.
!          
!
! iounit = logical unit number where to write the matrix into.
!
! on return:
!---------- 
! the matrix a, ja, ia will be written in output unit iounit
! in the Harwell-Boeing format. None of the inputs is modofied.
!  
! Notes: 1) This code attempts to pack as many elements as possible per
!        80-character line. 
!        2) this code attempts to avoid as much as possible to put
!        blanks in the formats that are written in the 4-line header
!    (This is done for purely esthetical reasons since blanks
!        are ignored in format descriptors.)
!        3) sparse formats for right hand sides and guesses are not
!        supported.
!-----------------------------------------------------------------------
      character title*72,key*8,rua*3,ptrfmt*16,indfmt*16,valfmt*20,    &
                guesol*2, rhstyp*3
      integer :: totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,   &
           nnz, nrhs, length, nperli
      integer :: ja(*), ia(*)  
      double precision :: a(*),rhs(*)
!--------------
!     compute pointer format
!--------------
      nnz    = ia(ncol+1) -1
      length    = int ( alog10(0.1+real(nnz+1))) + 1
      nperli = 80/length
      ptrcrd = ncol/nperli + 1
      if (length  >  9) then
         assign 101 to ix
      else
         assign 100 to ix
      endif
      write (ptrfmt,ix) nperli,length
 100  format(1h(,i2,1Hi,i1,1h) )
 101  format(1h(,i2,1Hi,i2,1h) )
!----------------------------
! compute ROW index format
!----------------------------
      length    = int ( alog10(0.1+real(nrow) )) + 1
      nperli = min0(80/length,nnz)
      indcrd = (nnz-1)/nperli+1
      write (indfmt,100) nperli,length
!---------------
! compute values and rhs format (using the same for both)
!--------------- 
      valcrd    = 0
      rhscrd  = 0
! quit this part if no values provided.
      if (job  <=  1) goto 20

      if (ifmt  >=  100) then
         ihead = ifmt/100
         ifmt = ifmt-100*ihead
         length = ihead+ifmt+2
         nperli = 80/length

         if (length  <=  9 ) then
            assign 102 to ix
         elseif (ifmt  <=  9) then
            assign 103 to ix
         else 
            assign 104 to ix
         endif

         write(valfmt,ix) nperli,length,ifmt
 102     format(1h(,i2,1hf,i1,1h.,i1,1h) )
 103     format(1h(,i2,1hf,i2,1h.,i1,1h) )
 104     format(1h(,i2,1hf,i2,1h.,i2,1h) )

      else
         length = ifmt + 6
         nperli = 80/length
!     try to minimize the blanks in the format strings.
         if (nperli  <=  9) then
        if (length  <=  9 ) then
           assign 105 to ix
        elseif (ifmt  <=  9) then
           assign 106 to ix
        else 
           assign 107 to ix
        endif
     else 
        if (length  <=  9 ) then
           assign 108 to ix
        elseif (ifmt  <=  9) then
           assign 109 to ix
        else 
               assign 110 to ix
            endif
         endif

         write(valfmt,ix) nperli,length,ifmt
 105     format(1h(,i1,1hd,i1,1h.,i1,1h) )
 106     format(1h(,i1,1hd,i2,1h.,i1,1h) )
 107     format(1h(,i1,1hd,i2,1h.,i2,1h) )
 108     format(1h(,i2,1hd,i1,1h.,i1,1h) )
 109     format(1h(,i2,1hd,i2,1h.,i1,1h) )
 110     format(1h(,i2,1hd,i2,1h.,i2,1h) )

      endif         
      valcrd = (nnz-1)/nperli+1
      nrhs   = job -2
      if (nrhs  >=  1) then
         i = (nrhs*nrow-1)/nperli+1
         rhscrd = i
         if (guesol(1:1)  ==  'g') guesol(1:1) = 'G'
         if (guesol(1:1)  ==  'G' .or. guesol(1:1)  ==  'g')    &
             rhscrd = rhscrd+i
         if (guesol(2:2)  ==  'x') guesol(2:2) = 'X'
         if (guesol(2:2)  ==  'X' .or. guesol(2:2)  ==  'x')    &
             rhscrd = rhscrd+i
         rhstyp = 'F'//guesol
      endif 
 20   continue

      totcrd = ptrcrd+indcrd+valcrd+rhscrd
!     write 4-line or five line header
      write(iounit,10) title,key,totcrd,ptrcrd,indcrd,valcrd,       &
           rhscrd,rua,nrow,ncol,nnz,nrhs,ptrfmt,indfmt,valfmt,valfmt
!-----------------------------------------------------------------------
      if (nrhs  >=  1) write (iounit,11) rhstyp, nrhs
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
 11   format(A3,11x,i4)

      write(iounit,ptrfmt) (ia (i), i = 1, ncol+1)
      write(iounit,indfmt) (ja (i), i = 1, nnz)
      if (job  <=  1) return
      write(iounit,valfmt) (a(i), i = 1, nnz)
      if (job  <=  2) return 
      length = nrow*nrhs 
      next = 1
      iend = length
      write(iounit,valfmt) (rhs(i), i = next, iend)
!     
!     write initial guesses if available
!     
      if (guesol(1:1)  ==  'G' .or. guesol(1:1)  ==  'g') then
         next = next+length
         iend = iend+ length
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!     
!     write exact solutions if available
!     
      if (guesol(2:2)  ==  'X' .or. guesol(2:2)  ==  'x')then
         next = next+length
         iend = iend+ length
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!     
      return
      end subroutine prtmt

!======================================================================
      subroutine descr(iunit)
!----------------------------------------------------------------------
!
!  This routine just dumps out pretty much the same thing as is in
!  the splib.F file documentation, but in a nifty HTML format
!  suitable for browsers or decorating wastebaskets.  Logical file
!  iunit is used to write to.
!     
!----------------------------------------------------------------------
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University, Bloomington
!     bramley@cs.indiana.edu
!     Mon Apr 12 18:27:31 EST 1999
!
!----------------------------------------------------------------------

      integer :: iunit
      logical ofl

!      inquire(iunit, opened=ofl)
!      if (opened) go to 5

      open(iunit,file='About_SPLIB.html',STATUS="REPLACE",err=5)
      rewind(iunit)

      write(iunit,2045 )
      write(iunit,8010 )
      write(iunit,8020 )
      write(iunit,8030 )
      write(iunit,8040 )
      write(iunit,8050 )
      write(iunit,8060 )
      write(iunit,8070 )
      write(iunit,8080 )
      write(iunit,8090 )
      write(iunit,8100 )
      write(iunit,8110 )
      write(iunit,8120 )
      write(iunit,8130 )
      write(iunit,8140 )
      write(iunit,8150 )
      write(iunit,8160 )
      write(iunit,8170 )
      write(iunit,8180 )
      write(iunit,8190 )
      write(iunit,8200 )
      write(iunit,8210 )
      write(iunit,8220 )
      write(iunit,8230 )
      write(iunit,8240 )
      write(iunit,8250 )
      write(iunit,8260 )
      write(iunit,8270 )
      write(iunit,8280 )
      write(iunit,8290 )
      write(iunit,8300 )
      write(iunit,8310 )
      write(iunit,8320 )
      write(iunit,8330 )
      write(iunit,8340 )
      write(iunit,8350 )
      write(iunit,8360 )
      write(iunit,8370 )
      write(iunit,8380 )
      write(iunit,8385 ) '(m)ilus', 'n', '2n'
      write(iunit,8385 ) '   ilut', '2n + 1', '3n'
      write(iunit,8385 ) 'ssor   ', '1     ','n' 
      write(iunit,8385 ) 'trid','4n','n'
      write(iunit,8385 ) 'ilu0 ','n','n'
      write(iunit,8385 ) 'ecimgs','?','?' 
      write(iunit,8385 ) 'bicg','6n', '0'
      write(iunit,8385 ) 'cgne','3n', '0'
      write(iunit,8385 ) 'cgnr','3n', '0'
      write(iunit,8385 ) 'cgs','6n', '0'
      write(iunit,8385 ) 'cgstab','6n', '0'
      write(iunit,8385 ) 'gmres(k)','k*(n+k+5)+n + 1', '0'
      write(iunit,8385 ) 'qmrtf','10n', '0'
      write(iunit,8385 ) 'bicgs','6n', '0'
      write(iunit,8385 ) 'orthomin(k)','(2k+6)*n', '0'
      write(iunit,8385 ) 'jacobi', '0', '0'
      write(iunit,8385 ) 'gauss', '0', '0'
      write(iunit,8385 ) 'SOR', '0', '0'
      write(iunit,8390 )
      write(iunit,8400 )
      write(iunit,8410 )
      write(iunit,8420 )
      write(iunit,2050 )

 5    continue

      return

!===========================================================================
!                 *** Format Statements ***
!===========================================================================

 2045  format('<HTML>',/, '<HEAD>',/,                               &
         '<A NAME="About"><TITLE>About Sparse Iterative Library',/, &
         ' (SPLIB)</TITLE></A>',/,                                  &
         '</HEAD>',/,'<BODY TEXT="#000000" BGCOLOR="#FDF5E6">',/,   &
         '<HR size=8 noshade><H1>About Sparse Iterative Library',/, &
         ' (SPLIB) </H1>')

 2050  format(/,'</BODY>',/,'</HTML>')

 8010  format('<HR size=10 noshade>',/,                     &
        'The purpose of SPLIB is to provide a',/,           &
        'collection of preconditioned iterative methods for',/,   &
        'practical testing',/,                              &
        'on systems of nontrivial size without',/,          &
        'massive amounts of coding',/,                      &
        'effort.  In particular, SPLIB relieves the',/,     &
        'user of the need to code',/,                       &
        'up preconditioners, which are often the most',/,   &
        'difficult and error-prone',/,                      &
        'part of iterative solvers.  The price to be',/,    &
        'paid is that the matrix',/,                        &
        'must be in a specific data structure:',/,          &
        'compressed sparse row.  This stores',/,            &
        'the nonzero entries of A in a double',/,           &
        'precision singly dimensioned vector,',/,           &
        'a corresponding integer vector of the column')
 8020  format('indices, and an integer',/,                  &
        'vector of length n+1 with pointers to the',/,      &
        'beginning of each row in the',/,                   &
        'matrix ( entry n+1 points to one position',/,      &
        'beyond the end of the arrays',/,                   &
        'for nonzero values and column indices).',/,        &
        '<P>',/,                                            &
        'The target users for SPLIB are scientists',/,      &
        'and engineers in application',/,                   &
        'areas who wish to experiment with iterative',/,    &
        'solvers in their codes to',/,                      &
        'find which of the many methods may be',/,          &
        'practical for them.  SPLIB is',/,                  &
        'also useful for teaching iterative methods,',/,    &
        'allowing students to',/,                           &
        'implement methods without having to develop',/,    &
        'the massive amounts of code',/,                    &
        'needed for instrumentation, etc.')
 8030  format('<P>',/,                                      &
        'The code is written in Fortran 77, with a',/,      &
        'few exceptions.',/,                                &
        'Include files are used for some parameters',/,     &
        'and the memory manager.',/,                        &
        'One function, getr.c, is necessarily written',/,   &
        'in C.  That',/,                                    &
        'function accesses the getrusage structure',/,      &
        'and returns it to the',/,                          &
        'Fortran 77 routine prtres.  A timing',/,           &
        'function (mytime)',/,                              &
        'is used.  It is set to use etime(), but has',/,    &
        'commented out options',/,                          &
        'for using mclock(), second(), etc. on',/,          &
        'machines without etime().',/,                      &
        'BLAS routines used include: ddot, daxpy,',/,       &
        'dcopy, dgemv,',/,                                  &
        'dasum, drot, dscal, drotg, and dtrsv.')
 8040  format('Routines lsame and xerbla',/,        &
        'from the LAPACK project are also provided',/,        &
        'in the directory "blas",',/,        &
        'so that SPLIB can be compiled and run on',/,        &
        'machines without a native',/,        &
        'BLAS library. "implicit none"',/,        &
        'is used in all the routines.',/,        &
        '<P>',/,        &
        'This package differs from others in seven',/,        &
        'ways:',/,        &
        '<OL>',/,        &
        '  <LI> It uses a sophisticated memory',/,        &
        '  managment module, which allows',/,        &
        '     the package to be easily adapted and',/,        &
        '     used by applications')
 8050  format('     programs.  All they need do is provide',/,        &
        '     a work array, from which',/,        &
        '     the necessary arrays are divvied out',/,        &
        '     using pointers.  This helps',/,        &
        '     keep the calling sequence short.',/,        &
        '  <LI> It provides implementations of levels',/,        &
        '  of fill LU preconditioning,',/,        &
        '     well documented and with various',/,        &
        '     options.',/,        &
        '  <LI> The matrices dealt with are always',/,        &
        '  kept in standard data structures,',/,        &
        '     so that they can be easily extracted',/,        &
        '     and examined/processed using',/,        &
        '     other tools, such as emily for',/,        &
        '     visualization or Sparskit for basic',/,        &
        '     manipulations.',/,        &
        '  <LI> Detailed instrumentation of the code')
 8060  format('  is provided for, at the risk of',/,        &
        '     losing some portability.  In',/,        &
        '     particular, this code allows tracking',/,        &
        '     several different measures of error',/,        &
        '     norms, including Oetli backward',/,        &
        '     error analysis based measure.',/,        &
        '     Furthermore, the error norms are',/,        &
        '     written',/,        &
        '     out to files which can then be plotted',/,        &
        '     via xgraph, giving graphical',/,        &
        '     comparison of different methods and',/,        &
        '     preconditioners.  Finally, the',/,        &
        '     norms are plotted against both',/,        &
        '     iteration number and time, allowing',/,        &
        '     a more fair comparison than the usual',/,        &
        '     "number of iterations" results.',/,        &
        '     The plotting against time is especially',/,        &
        '     crucial, since methods can')
 8070  format('     vary drastically in the amount of work',/,        &
        '     per iteration.',/,        &
        '  <LI> Further instrumentation provides',/,        &
        '  access to rusage structures, which',/,        &
        '     allows information about memory',/,        &
        '     utilization, page faults, context',/,        &
        '     switches, and other important',/,        &
        '     performance measures.',/,        &
        '  <LI> Implementation of the partial',/,        &
        '  orthogonalization preconditioners of',/,        &
        '     Bramley, Gallivan, and Wang.',/,        &
        '     These generally produce better factors',/,        &
        '     than the ILU type methods, and are',/,        &
        '     more robust.  However, they require',/,        &
        '     more intermediate storage to compute.',/,        &
        '  <LI> Stringent tests for having too large',/,        &
        '  of a preconditioned')
 8080  format('     residual, for dividing by small',/,        &
        '     quantities, etc.',/,        &
        '</OL>',/,        &
        '',/,        &
        '<H3>History of SPLIB:</H3>',/,        &
        'This code has evolved through many hands.',/,        &
        'Originally it started',/,        &
        'out as RP-PACK, for examining row projection',/,        &
        'methods.  It then',/,        &
        'was converted to a front end for CG-like',/,        &
        'iterative methods.',/,        &
        'Ulrike Meier and G. Skinner did most of the',/,        &
        'preliminary work, for',/,        &
        'the Center for Supercomputer Research and',/,        &
        'Development at the University',/,        &
        'of Illinois.  They created a uniform calling',/,        &
        'sequence and data naming')
 8090  format('convention that tied many disparate codes',/,        &
        'together, as well as writing and',/,        &
        'testing heroic amounts of codes.  Xiaoge',/,        &
        'Wang has done a great deal',/,        &
        'of work, especially on the incomplete',/,        &
        'preconditioners based on levels',/,        &
        'of fill-in, and all of the ECIMGS material.',/,        &
        '<P>',/,        &
        'Youcef Saad wrote the original GMRES(k)',/,        &
        'solver, and the package',/,        &
        'heavily uses modified SPARSKIT routines',/,        &
        'provided by him and Edmond Chow.')
 8100  format('',/,        &
        '<H2><A NAME="Usage">Usage of SPLIB</a></H2>',/,        &
        'Compilation is via the makefile.',/,        &
        'Note: You should first edit the include file',/,        &
        'intdbl.inc, to give the',/,        &
        '    number of integer words per double',/,        &
        '    precision word.  For most',/,        &
        '    machines this is intdbl = 2, but for',/,        &
        '    Cray machines it is',/,        &
        '    generally intdbl = 1.  For 64 bit',/,        &
        '    machines with long',/,        &
        '    integers, this is also intdbl = 1.',/,        &
        '    This is used to force',/,        &
        '    double word alignment in the memory',/,        &
        '    manager. <BR>')
 8110  format( 'Compilation sets up the archive library',/,        &
        'splib.a, which you can',/,        &
        'link with your code.',/,        &
        '<P>',/,        &
        'Calling SPLIB requires you to:',/,        &
        '<OL>',/,        &
        '<LI>  Set up the matrix in CSR format.',/,        &
        '    The main sticking point in using SPLIB')
 8120  format('    is that the matrix must be',/,        &
        '    provided in compressed sparse row',/,        &
        '    (CSR) format.  This is necessary',/,        &
        '    because the preconditioners need to be',/,        &
        '    able to access the',/,        &
        '    entries of the matrix.  Unpreconditioned',/,        &
        '    iterative methods based on',/,        &
        '    Krylov subspaces generally only need to',/,        &
        '    access the matrix via',/,        &
        '    matrix-vector products, but they also',/,        &
        '    generally do not work.  The',/,        &
        '    tradeoff here is that most applications',/,        &
        '    scientists find it relatively',/,        &
        '    easy to write a data structure',/,        &
        '    conversion routine to put their',/,        &
        '    coefficient matrix into a CSR format,',/,        &
        '    but writing robust factorization',/,        &
        '    based preconditioners are much more')
 8130  format('    difficult.',/,        &
        '<LI>  Provide n-vectors x and b for the',/,        &
        'solution and right hand side: A x = b.',/,        &
        '<LI>  Provide an integer array',/,        &
        'work(1:space), aligned on a double word',/,        &
        '    boundary.  All of the additional storage',/,        &
        '    needed by SPLIB will be',/,        &
        '    carved out from this array.  To assure',/,        &
        '    double word alignment, you can',/,        &
        '    declare it as a double precision array',/,        &
        '    of length m, then pass it in',/,        &
        '    with space = intdbl*m.',/,        &
        '<LI>  Set the control parameters.  Defaults',/,        &
        'are in parenthesis, and are',/,        &
        '    triggered by giving an invalid value',/,        &
        '    for the parameter.',/,        &
        '<DL>',/,        &
        '<DT> <B>solmeth</B>')
 8140  format('   <DD> solution method',/,        &
        '     (CG-Stabilized, solmeth = 5)',/,        &
        '<DT> <B>pcmeth</B>',/,        &
        '     <DD> preconditioning method (No',/,        &
        '     preconditioning, pcmeth = 0)',/,        &
        '<DT> <B>iredo</B>',/,        &
        '     <DD> whether or not to recompute',/,        &
        '     the preconditioner.',/,        &
        '         (0 for recompute, 1',/,        &
        '         to not).',/,        &
        '<DT> <B>tol</B>',/,        &
        '     <DD> tolerance on residual',/,        &
        '     (1.0e-6)',/,        &
        '<DT> <B>ireltv</B>',/,        &
        '     <DD> whether or not to use',/,        &
        '     relative residual for stopping',/,        &
        '         (0 for not, 1 to use')
 8150  format('       relative residual).',/,        &
        '<DT> <B>maxits</B>',/,        &
        '     <DD> maximal allowed iterations',/,        &
        '     (100)',/,        &
        '<DT> <B>scale</B>',/,        &
        '     <DD> scaling of the system (0)',/,        &
        '<DT> <B>instlvl</B>',/,        &
        '     <DD> instrumentation level (0)',/,        &
        '</DL>',/,        &
        '    In addition, some methods require',/,        &
        '    setting <B>iparms</B> and <B>rparms</B>. For',/,        &
        '    example, GMRES(m) requires a value m',/,        &
        '    for the Krylov subspace size (but',/,        &
        '    this is set to 10 by default).  ILU(s)',/,        &
        '    factorization requires a level',/,        &
        '    of fill value, which by default is 0.',/,        &
        '    Because of the wide variety of')
 8160  format('    methods available, it is necessary to',/,        &
        '    check the individual routines',/,        &
        '    to find all the possibilities.',/,        &
        '<LI>  If instrumentation is to be used,',/,        &
        'a vector of 17 logical unit numbers',/,        &
        '    must be provided.  SPLIB will open files',/,        &
        '    with those unit numbers, to',/,        &
        '    write out summary data and Xgraph files',/,        &
        '    tracing the iteration history.',/,        &
        '</OL>',/,        &
        '',/,        &
        '<H2><A NAME="Methods">Methods and their',/,        &
        'specification</A></H2>',/,        &
        'The two major choices are the iterative',/,        &
        'solver and the',/,        &
        'preconditioner to use. Those are specified',/,        &
        'by solmeth and',/,        &
        'pcmeth, with the following interpretation:',/,        &
        '<DL COMPACT>')
 8170  format('<DT> <B>solmeth</B> = 1',/,        &
        '      <DD> BiCG, biconjugate gradients',/,        &
        '<DT> <B>solmeth</B> = 2',/,        &
        '      <DD> CG on normal equations AA"y',/,        &
        '     </B> = b; x = A"y.',/,        &
        '<DT> <B>solmeth</B> = 3',/,        &
        '      <DD> CG on normal equations A"A',/,        &
        '      x</B> = A"b',/,        &
        '<DT> <B>solmeth</B> = 4',/,        &
        '      <DD> CGS, conjugate gradients',/,        &
        '      squared',/,        &
        '<DT> <B>solmeth</B> = 5',/,        &
        '      <DD> CGSTAB, stabilized version',/,        &
        '      of conjugate gradients squared',/,        &
        '<DT> <B>solmeth</B> = 6',/,        &
        '      <DD> GMRES(k), generalized',/,        &
        '      minimum residual',/,        &
        '<DT> <B>solmeth</B> = 7')
 8180  format('    <DD> transpose-free QMR',/,        &
        '<DT> <B>solmeth</B> = 8',/,        &
        '      <DD> Bi-CG stabilized (Templates',/,        &
        '      version of CG-Stab).',/,        &
        '<DT> <B>solmeth</B> = 9',/,        &
        '      <DD> Templates version of GMRES',/,        &
        '<DT> <B>solmeth</B> = 10',/,        &
        '      <DD> Jacobi',/,        &
        '<DT> <B>solmeth</B> = 11',/,        &
        '      <DD> Gauss-Seidel',/,        &
        '<DT> <B>solmeth</B> = 12',/,        &
        '      <DD> SOR, successive',/,        &
        '      over-relaxation',/,        &
        '<DT> <B>solmeth</B> = 13',/,        &
        '      <DD> Orthomin(m)',/,        &
        '</DL>',/,        &
        '',/,        &
        'For preconditioners, the interpretations')
 8190  format('follow.  Beware that not',/,        &
        'all iterative solvers can be used with all',/,        &
        'preconditioners; for',/,        &
        'example, the Jacobi, Gauss-Seidel, and SOR',/,        &
        'cannot be used with',/,        &
        'the (M)ILU preconditioners.',/,        &
        '<DL COMPACT>',/,        &
        '<DT> <B>pcmeth</B> = 0',/,        &
        '      <DD> no preconditioner.',/,        &
        '<DT> <B>pcmeth</B> = 1',/,        &
        '      <DD> ILU(s), incomplete LU',/,        &
        '      factorization with s levels',/,        &
        '      of fill',/,        &
        '<DT> <B>pcmeth</B> = 2',/,        &
        '      <DD> MILU(s,r), modifed',/,        &
        '      incomplete LU factorization with',/,        &
        '       s levels of fill and',/,        &
        '       modification factor r')
 8200  format('<DT> <B>pcmeth</B> = 3',/,        &
        '      <DD> ILUT(s,r), Saads ILU with',/,        &
        '      drop tolerance r',/,        &
        '<DT> <B>pcmeth</B> = 4',/,        &
        '      <DD> SSOR(r), symmetric',/,        &
        '      successive overrelaxation',/,        &
        '<DT> <B>pcmeth</B> = 5',/,        &
        '      <DD> TRID(s), where s is the',/,        &
        '      block size.',/,        &
        '<DT> <B>pcmeth</B> = 6',/,        &
        '      <DD> ILU(-1), the space saver',/,        &
        '      version of ILU(0)',/,        &
        '<DT> <B>pcmeth</B> = 7',/,        &
        '      <DD> ECIMGS, extended compressed',/,        &
        '      incomplete modified Gram-Schmidt',/,        &
        '</DL>',/,        &
        '',/,        &
        '<H2><A NAME="Input">Arguments to')
 8210  format('SPLIB</A></H2>',/,        &
        '<UL>',/,        &
        '<LI><B>n</B>: Order of the matrix A.',/,        &
        '<LI><B>A</B>: double precision :: array for nonzero',/,        &
        'entries of matrix A; of',/,        &
        '     length nnz = rwptr(n+1)-1.',/,        &
        '<LI><B>colind</B>: Integer array of column',/,        &
        'indices for entries in A; of',/,        &
        '     length nnz = rwptr(n+1)-1.',/,        &
        '<LI><B>rwptr</B>: Integer array of pointers',/,        &
        'to beginning position of rows',/,        &
        '     in arrays A and colind.  Must be',/,        &
        '     of length n+1.',/,        &
        '',/,        &
        '<LI><B>x</B>: double precision :: vector of unknowns to',/,        &
        'be solved for.',/,        &
        '<LI><B>b</B>: double precision :: right hand side vector',/,        &
        'of system to be solved: Ax = b.')
 8220  format('   See lrhs and lsol for options on',/,        &
        '     this vector.',/,        &
        '',/,        &
        '<LI><B>tol</B>: tolerance on norm used in',/,        &
        'stopping test.',/,        &
        '<LI><B>maxits</B>: Maximum number of',/,        &
        'iterations allowed.  On return, contains',/,        &
        'actual number performed.',/,        &
        '',/,        &
        '<LI><B>solmeth</B>: Solution method to use.',/,        &
        '',/,        &
        '<LI><B>pcmeth</B>: Preconditioning method',/,        &
        'to use.',/,        &
        '',/,        &
        '<LI><B>rparms</B>: real preconditioner',/,        &
        'parameter,',/,        &
        '  with interpretation depending on',/,        &
        '  preconditioning method to be used.')
 8230  format('    <DL COMPACT>',/,        &
        '      <DT>ILUT <DD> tolerance for',/,        &
        '      numerical dropping',/,        &
        '      <DT>SSOR <DD> omega parameter',/,        &
        '      for SSOR(omega)',/,        &
        '      <DT>MILU <DD> relaxation constant',/,        &
        '      <DT>ILU0 <DD> relaxation constant',/,        &
        '      <DT>ECIMGS  <DD> sparsity pattern',/,        &
        '      indicator for ECIMGS',/,        &
        '         <DL COMPACT>',/,        &
        '        <DT>rparms = 0.0',/,        &
        '        <DD>Use nonzero pattern of',/,        &
        '        A for L/U',/,        &
        '        <DT>rparms = 1.0',/,        &
        '        <DD>Use a diagonal matrix',/,        &
        '        as precond.',/,        &
        '        <DT>0.0 &lt rparms &lt 1.0',/,        &
        '        <DD>Use a pattern sparser')
 8240  format('      than A"s.',/,        &
        '        <DT>rparms &gt 1.0',/,        &
        '        <DD>Use int(rparms) levels',/,        &
        '        of fill-in.',/,        &
        '         </DL>',/,        &
        '      </DL>',/,        &
        '',/,        &
        '<LI><B>iparms</B>: integer parameters,',/,        &
        '  with interpretation depending on methods',/,        &
        '  used.',/,        &
        '      <DL COMPACT>',/,        &
        '      <DT> iparms(1) <DD> levels of',/,        &
        '      fill for ILU(s), MILU(s), ILUT',/,        &
        '      <DT> iparms(1) <DD> block size',/,        &
        '      for TRID.',/,        &
        '      <DT> iparms(2) <DD> maximum',/,        &
        '      Kyrlov subspace size k',/,        &
        '              for GMRES(k),')
 8250  format('            Orthomin(k).',/,        &
        '      </DL>',/,        &
        '',/,        &
        '<LI><B>instlvl</B>: Level of',/,        &
        'instrumentation.  The level is cummulative,',/,        &
        '   so specifying output of Oetli/Prager',/,        &
        '   norms implies residual norms, etc.',/,        &
        '      <DL COMPACT>',/,        &
        '      <DT> instlvl &lt 0  <DD> no',/,        &
        '      output from splib',/,        &
        '      <DT> instlvl = 0  <DD> only',/,        &
        '      summary data',/,        &
        '      <DT> instlvl = 1  <DD> residual',/,        &
        '      norm data',/,        &
        '      <DT> instlvl = 2  <DD>',/,        &
        '      preconditioned residual norm',/,        &
        '      <DT> instlvl = 3  <DD> relative',/,        &
        '      residual norm')
 8260  format('    <DT> instlvl = 4  <DD> relative',/,        &
        '      preconditioned residual norm',/,        &
        '      <DT> instlvl = 5  <DD>',/,        &
        '      Oetli/Prager norms',/,        &
        '      <DT> instlvl = 6  <DD> error',/,        &
        '      norms (if true solution vector',/,        &
        '                 is',/,        &
        '                 provided',/,        &
        '                 via a',/,        &
        '                 common',/,        &
        '                 block)',/,        &
        '      </DL>',/,        &
        '',/,        &
        '<LI><B>iunit</B>: vector of I/O unit numbers',/,        &
        'to use for output.',/,        &
        '     The I/O unit numbers are provided',/,        &
        '     in the vector iunit(17).',/,        &
        '     The units are optional, depending')
 8270  format('   on the error norm(s) you wish to',/,        &
        '     plot/use.  However, if you specify',/,        &
        '     instlvl >= 0,',/,        &
        '     the following units will be opened',/,        &
        '     for writing:',/,        &
        '     <DL COMPACT>',/,        &
        '     <DT> iunit(1) <DD>    File for',/,        &
        '     output of summary data',/,        &
        '     <DT> iunit(2-3) <DD>  Not used',/,        &
        '     here',/,        &
        '     <DT> iunit(4) <DD>    || Ax - b ||',/,        &
        '     vs. iter',/,        &
        '     <DT> iunit(5) <DD>    || inv(M)',/,        &
        '     (Ax-b) || vs. iter',/,        &
        '     <DT> iunit(6) <DD>    || Ax - b ||',/,        &
        '     / || r<SUB>0</SUB> || vs. iter',/,        &
        '     <DT> iunit(7) <DD>    ||',/,        &
        '     inv(M) (Ax-b) || / || inv(M)')
 8280  format('   r<SUB>0</SUB> || vs. iter',/,        &
        '     <DT> iunit(8) <DD> || Ax - b',/,        &
        '     ||<SUB>inf</SUB> /',/,        &
        '       ( ||A||<SUB>inf</SUB>',/,        &
        '       ||x||<SUB>1</SUB> +',/,        &
        '       ||b||<SUB>inf</SUB>) vs. iter',/,        &
        '     <DT> iunit(9) <DD>  || x - xstar',/,        &
        '     || vs. iter',/,        &
        '     <DT> iunit(10) <DD> Not used.',/,        &
        '     <DT> iunit(11) <DD> || Ax - b ||',/,        &
        '     vs. time',/,        &
        '     <DT> iunit(12) <DD> || inv(M)',/,        &
        '     (Ax-b) || vs. time',/,        &
        '     <DT> iunit(13) <DD> || Ax - b ||',/,        &
        '     / || r<SUB>0</SUB> || vs. time',/,        &
        '     <DT> iunit(14) <DD> ||',/,        &
        '     inv(M) (Ax-b) || / || inv(M)',/,        &
        '     r<SUB>0</SUB> ||')
 8290  format('            vs. time',/,        &
        '     <DT> iunit(15) <DD> || Ax -',/,        &
        '     b ||<SUB>inf</SUB> /',/,        &
        '        ( ||A||<SUB>inf</SUB>',/,        &
        '        ||x||<SUB>1</SUB> +',/,        &
        '        ||b||<SUB>inf</SUB>)',/,        &
        '              vs. time',/,        &
        '     <DT> iunit(16) <DD> || x - xstar',/,        &
        '     || vs. time',/,        &
        '      </DL>',/,        &
        '     <DT> iunit(17) <DD> This explanatory file',/,        &
        '<LI><B>work</B>: Array of work storage,',/,        &
        'for preconditioners, solvers, and.',/,        &
        '     instrumentation routines.',/,        &
        '     The matrices and vectors used',/,        &
        '     are all indexed from this integer',/,        &
        '     array, using pointers.',/,        &
        '')
 8300  format('<LI><B>space</B>: maximum number of integer',/,        &
        'words allowed for all problem',/,        &
        '     dependent arrays.  This includes',/,        &
        '     the preconditioners',/,        &
        '     and the additional work vectors',/,        &
        '     needed in the solvers.',/,        &
        '     On return, space holds the actual',/,        &
        '     amount of storage',/,        &
        '     required by SPLIB.',/,        &
        '',/,        &
        '<LI><B>iredo</B>: Recompute preconditioner',/,        &
        'if redo = 0, reuse if redo = 1.',/,        &
        '     This is useful if you wish to',/,        &
        '     reuse a preconditioner from',/,        &
        '     an earlier call to SPLIB.',/,        &
        '     Converted internally to logical',/,        &
        '     variable redo',/,        &
        '')
 8310  format('<LI><B>err</B>: Final error norm (2-norm)',/,        &
        '<LI><B>bnorm</B>: Two-norm of the right hand',/,        &
        'side vector b. Used for relative',/,        &
        '     residual stopping tests:',/,        &
        '     ||r<SUB>k</SUB>||/||r<SUB>0</SUB>||',/,        &
        '     < tol.',/,        &
        '<LI><B>sptime</B>: Integer pointer to array',/,        &
        'of times in memory manager.',/,        &
        '',/,        &
        '<LI><B>scale</B>: Job control parameter for',/,        &
        'scaling of system.',/,        &
        '   <DL COMPACT>',/,        &
        '      <DT> scale = 0  <DD> perform',/,        &
        '      no scaling',/,        &
        '      <DT> scale = 1  <DD> set',/,        &
        '      max(max(A)) = 1.0',/,        &
        '      <DT> scale = 2  <DD> set max of',/,        &
        '      each row to be 1.0')
 8320  format('    <DT> scale = 3  <DD> scale each',/,        &
        '      column so diagonal entry is 1.0',/,        &
        '      <DT> scale = 4  <DD> perform',/,        &
        '      first scale=2, then scale=3',/,        &
        '   </DL>',/,        &
        'Note that',/,        &
        '   <UL>',/,        &
        '     <LI> If scale = 0 or 1, Dl and Dr',/,        &
        '     can be dummy variables.',/,        &
        '     <LI> If scale = 2 or 4, Dl must',/,        &
        '     provided (see below)',/,        &
        '     <LI> If scale = 3 or 4, Dr must',/,        &
        '     provided (see below)',/,        &
        '   </UL>',/,        &
        'and that SPLIB will unscale the solution',/,        &
        'vector so that',/,        &
        'it is the solution to the original system',/,        &
        'when')
 8330  format('scale = 3 or 4.',/,        &
        '<LI><B>Dl, Dr</B>: Double arrays of length',/,        &
        'n, used to hold scaling of matrix.',/,        &
        '     system actually solved is Dl*A*Dr',/,        &
        '     y = Dl b.  On return, the',/,        &
        '     matrix A has been scaled and x',/,        &
        '     = Dr*y.',/,        &
        '</UL>',/,        &
        '',/,        &
        '<H2><A NAME="IO">I/O files</A></H2>',/,        &
        '<DL COMPACT>',/,        &
        '<DT> <B>results</B><DD> File containing',/,        &
        'summary of results from run.',/,        &
        '',/,        &
        '<DT> <B>resi</B>   <DD> File with residual',/,        &
        'norm vs. iteration number.',/,        &
        '<DT> <B>rest</B>   <DD> File with residual')
 8340  format('norm vs. elapsed time.',/,        &
        '',/,        &
        '<DT> <B>reli</B>   <DD> File with relative',/,        &
        'residual norm vs. iteration number.',/,        &
        '<DT> <B>relt</B>   <DD> File with relative',/,        &
        'residual norm vs. elapsed time.',/,        &
        '',/,        &
        '<DT> <B>Mresi</B>  <DD> File with',/,        &
        'preconditioned residual norm vs. iteration',/,        &
        'number.',/,        &
        '<DT> <B>Mrest</B>  <DD> File with',/,        &
        'preconditioned residual norm vs. elapsed',/,        &
        'time.',/,        &
        '',/,        &
        '<DT> <B>Mreli</B>  <DD> File with',/,        &
        'preconditioned relative residual norm',/,        &
        '     vs. iteration number.',/,        &
        '<DT> <B>Mrelt</B>  <DD> File with')
 8350  format('preconditioned relative residual norm',/,        &
        '     vs. elapsed time.',/,        &
        '',/,        &
        '<DT> <B>oetli</B>  <DD> File with Oettli',/,        &
        'and Prager norm vs. iteration number.',/,        &
        '<DT> <B>oetlt</B>  <DD> File with Oettli',/,        &
        'and Prager norm vs. elapsed time.',/,        &
        '',/,        &
        '<DT> <B>errori</B> <DD> File with error norm',/,        &
        'vs. iteration number',/,        &
        '              (if lsol = true).',/,        &
        '<DT> <B>errort</B> <DD> File with error norm',/,        &
        'vs. time',/,        &
        '              (if lsol = true).',/,        &
        '<DT> <B>About_SPLIB.html</B> <DD> This file!',/,        &
        '</DL>',/,        &
        '',/,        &
        '<H2><A NAME="Memory">Memory used in',/,        &
        'Splib</A></H2>')
 8360  format('The amount of workspace depends on the',/,        &
        'methods used, the levels',/,        &
        'of fill-in (and way of defining fill-in',/,        &
        'levels), and the number',/,        &
        'of Krylov subspace vectors used.',/,        &
        'The following guide excludes',/,        &
        'the storage for x, r, and b which appear in',/,        &
        'the calling',/,        &
        'sequence for all the routines. Note that',/,        &
        'GMRES(k), GMREST(k), and',/,        &
        'ORTHOMIN(k) depend on the size k of the',/,        &
        'Krylov subspace.  Also note',/,        &
        'that the methods with fill-in will require',/,        &
        'more storage for lu, jlu,',/,        &
        'uptr, but the actual amount is not',/,        &
        'predictable in advance.  The size',/,        &
        'shown is for the preconditioners with no',/,        &
        'fill-in allowed, and does')

 8370  format('not include the storage that depends on nnz.',/,        &
        'Even for zero levels of fill, ECIMGS requires intermediate',/,        &
        'storage that is wildly unpredictable.',/,        &
        'WARNING: for all of the methods, add 2n doubles if the',/,        &
        'instrumentation routine is used at all. <P><BR>')

 8380 format('<TABLE BORDER=8 WIDTH=80%>',/,'<CAPTION><B>',/,        &
        'Memory Usage Estimates (a priori)',/,'</B></CAPTION>',/,        &
        '<TR ALIGN="center">',/,'  <TH>Routine</TH><TH>Doubles</TH>',/,        &
        '  <TH>Integers</TH> </TR>')

 8385 format('<TR ALIGN="right">',/,        &
                 '<TD>',a12,'</TD>',        &
                 '<TD>',a15,'</TD>',        &
                 '<TD>',a3,'</TD>',        &
                 '</TR>')

 8390  format('</TABLE>',/,/,        &
        '<H2><A NAME="Error">Error flags',/,        &
        'returned</A></H2>',/,        &
        'Error conditions are specified in the',/,        &
        'integer variable errflg,',/,        &
        'with the following interpretations:',/,        &
        '<DL COMPACT>',/,        &
        '   <DT> <B>errflg = -m</B> <DD> Not enough',/,        &
        '   workspace provided; m more double',/,        &
        '              precision words',/,        &
        '              needed in array',/,        &
        '              work().',/,        &
        '   <DT> <B>errflg = 0</B>  <DD> Successful',/,        &
        '   return; tolerance achieved.',/,        &
        '   <DT> <B>errflg = 1</B>  <DD> The initial')

 8400  format('   guess satisfies the stopping test.',/,        &
        '              This is often',/,        &
        '              a clue that rhs',/,        &
        '              not set properly.',/,        &
        '   <DT> <B>errflg = 2</B>  <DD> Tolerance',/,        &
        '   reset 8 times in attempt to',/,        &
        '              get true residual',/,        &
        '              small enough.',/,        &
        '   <DT> <B>errflg = 3</B>  <DD> Convergence',/,        &
        '   not achieved in maxits iterations.',/,        &
        '   <DT> <B>errflg = 4</B>  <DD>',/,        &
        '   Preconditioned residual is too large',/,        &
        '   relative',/,        &
        '              to the',/,        &
        '              unpreconditioned',/,        &
        '              residual.',/,        &
        '              Indicates',/,        &
        '              an unstable')
 8410  format('            preconditioner.',/,        &
        '   <DT> <B>4 &lt errflg &lt 9</B> <DD>',/,        &
        '   Breakdown condition,',/,        &
        '                usually from',/,        &
        '                an attempt to',/,        &
        '                 divide by a',/,        &
        '                 near-zero',/,        &
        '                 number in',/,        &
        '                 algorithm.',/,        &
        '   <DT> <B>errflg = 10+k</B> <DD> Error condition',/,        &
        '   k occured in preconditioner.',/,        &
        '   <DT> <B>errflg = 100+k</B> <DD> Oetli-Prager measure',/,        &
        '   is less than machine tolerance but the residual is',/,        &
        '   still large.  Indicates have solved problem to ',/,        &
        '   within machine tolerance - it''s ill conditioned.',/,        &
        '</DL>',/,        &
        'However, note that any of conditions 0, 1,',/,        &
        '2 may mean success.')
 8420  format('This is up to the user to decide.')

        end
