!======================================================================== 
!                    **** utilsm.f *****
!========================================================================        
!
!  utilsm.f contains routines and functions that are generally useful for 
!  direct sparse matrix manipulation.
!  Currently contains the routines:
!
!      bmux    : CSR A - multiply: y = A*x
!      btmux   : CSR A'- multiply: y = A'*x
!      btmuxp  : CSR A'- multiply: y = y + A'*x
!      btmuxm  : CSR A'- multiply: y = y - A'*x
!      spbmux  : CSR A - multiply using sparse dotproduct function calls.
!      scala   : Scaling of matrix A and right hand side b
!      unscalx : Unscale solution x at end of computations.
!
!     Randall Bramley
!     Department of Computer Science
!     Indiana University
!     Phone: (812) 855-7790
!     email: bramley@cs.indiana.edu
!     Mon Dec 16 11:31:50 EST 1996
!
!
!========================================================================        

      subroutine bmux (m, n, x, y, a, ja, ia) 
!-----------------------------------------------------------------------
!          y = A*x
!----------------------------------------------------------------------- 
! Multiplies a matrix times a vector using the dot product form
! A is stored in compressed sparse row storage.  This version is 
! instrumented, by adding its total time to the times() array on
! each invocation.  
!
! on entry:
!----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the column dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
!-----------
! y     = real array of length m, containing the product y=Ax
!
!-----------------------------------------------------------------------
     
      implicit none

!     Arguments
!     =========
      integer m, n, ja(*), ia(*)
      double precision ::  x(n), y(m), a(*) 

!     Local variables
!     ===============
      double precision :: t
      integer i, k

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start Executable Statements
!=======================================================================

      tstart = mytime()

      do 100 i = 1, m

         y(i) = 0.0d0
!        ------------------------------------------------
!        Compute the inner product of row i with vector x
!        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue

      times(3) = times(3) + mytime() - tstart
      return

      end subroutine bmux

!=======================================================================
      subroutine btmux (m, n, x, y, a, ja, ia)
!-----------------------------------------------------------------------
!         y =  A' * x
!
! Multiplies the transpose of a matrix times a vector when the original
! matrix is stored in compressed sparse row storage. Can also be
! used as the product of a matrix by a vector when the original
! matrix is stored in the compressed sparse column format.
!-----------------------------------------------------------------------
!
! on entry:
! ----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the row dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
! -----------
! y     = real array of length equal to column dimmension of the A matrix, 
!	  containing the product y=transp(A)*x
!
!-----------------------------------------------------------------------

      implicit none

!     Arguments
!     =========
      integer m, n, ia(*), ja(*)
      double precision :: x(m), y(n), a(*)

!     Local variables
!     ===============
      integer i, k

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start of Executable Statements
!=======================================================================

      tstart = mytime()

!     ----------------------
!     Zero out output vector
!     ----------------------
      do i = 1,n
         y(i) = 0.0
      end do

!     -------------------------------------------------
!     Loop over the rows, using saxpy form of multiply.
!     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

      end subroutine btmux

      subroutine bmuxp(m, n, x, y, a, ja, ia) 
!-----------------------------------------------------------------------
!          y = y + A*x
!----------------------------------------------------------------------- 
! Multiplies a matrix times a vector using the dot product form
! A is stored in compressed sparse row storage.  This version is 
! instrumented, by adding its total time to the times() array on
! each invocation.  
!
! on entry:
!----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the column dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
!-----------
! y     = real array of length m, containing the product y=Ax
!
!-----------------------------------------------------------------------
     
      implicit none

!     Arguments
!     =========
      integer m, n, ja(*), ia(*)
      double precision ::  x(n), y(m), a(*) 

!     Local variables
!     ===============
      double precision :: t
      integer i, k

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start Executable Statements
!=======================================================================

      tstart = mytime()

      do 100 i = 1, m

!        ------------------------------------------------
!        Compute the inner product of row i with vector x
!        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue

      times(3) = times(3) + mytime() - tstart
      return

      end subroutine bmuxp
!
!=======================================================================
      subroutine btmuxp(m, n, x, y, a, ja, ia)
!-----------------------------------------------------------------------
!         y =  y + A' * x
!
! Multiplies the transpose of a matrix times a vector when the original
! matrix is stored in compressed sparse row storage. Can also be
! used as the product of a matrix by a vector when the original
! matrix is stored in the compressed sparse column format.  This version
! updates with an add to y.
!-----------------------------------------------------------------------
!
! on entry:
! ----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the row dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
! -----------
! y     = real array of length equal to column dimmension of the A matrix, 
!	  containing the product y=transp(A)*x
!
!-----------------------------------------------------------------------

      implicit none

!     Arguments
!     =========
      integer m, n, ia(*), ja(*)
      double precision :: x(m), y(n), a(*)

!     Local variables
!     ===============
      integer i, k

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start of Executable Statements
!=======================================================================

      tstart = mytime()

!     -------------------------------------------------
!     Loop over the rows, using saxpy form of multiply.
!     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

      end subroutine btmuxp
 
!=======================================================================
      subroutine btmuxm(m, n, x, y, a, ja, ia)
!-----------------------------------------------------------------------
!         y =  y - A' * x
!
! Multiplies the transpose of a matrix times a vector when the original
! matrix is stored in compressed sparse row storage. Can also be
! used as the product of a matrix by a vector when the original
! matrix is stored in the compressed sparse column format.  This version
! updates with a subtraction from y.
!-----------------------------------------------------------------------
!
! on entry:
! ----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the row dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
! -----------
! y     = real array of length equal to column dimmension of the A matrix, 
!	  containing the product y=transp(A)*x
!
!-----------------------------------------------------------------------

      implicit none

!     Arguments
!     =========
      integer m, n, ia(*), ja(*)
      double precision :: x(m), y(n), a(*)

!     Local variables
!     ===============
      integer i, k

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start of Executable Statements
!=======================================================================

      tstart = mytime()

!     -------------------------------------------------
!     Loop over the rows, using saxpy form of multiply.
!     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) - x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

      end subroutine btmuxm

!=======================================================================
      subroutine spbmux (m, n, x, y, a, ja, ia) 
!-----------------------------------------------------------------------
!          y = A*x
!----------------------------------------------------------------------- 
! Multiplies a matrix times a vector using the dot product form.
! A is stored in compressed sparse row storage.  This version is 
! instrumented, and it calls the C routine spdot.
!
! on entry:
!----------
! m     = row dimension of A
! n     = column dimension of A
! x     = real array of length equal to the column dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
!-----------
! y     = real array of length m, containing the product y=Ax
!
!-----------------------------------------------------------------------
     
      implicit none

!     Arguments
!     =========
      integer m, n, ja(*), ia(*)
      double precision ::  x(n), y(m), a(*) 
      double precision ::  spdot
      external spdot

!     Local variables
!     ===============
      double precision :: t
      integer i, k, start

!     Instrumentation variables for timing.
!     =====================================
      include '../include/heads.inc'
      double precision :: mytime, tstart
      external mytime

!=======================================================================
!      Start Executable Statements
!=======================================================================

      tstart = mytime()

      do 100 i = 1, m

!        ------------------------------------------------
!        Compute the inner product of row i with vector x
!        ------------------------------------------------
         start = ia(i)
         k = ia(i+1) - start
         y(i) = spdot(k, x, a(start), ja(start))

 100  continue
!
      times(3) = times(3) + mytime() - tstart
      return

      end subroutine spbmux

!=======================================================================

      subroutine scala(job,A, rwptr, colind, n, b, D)
      implicit none

!----------------------------------------------------------------
!     Scale a CSR matrix to have certain properties, 
!     depending on the job parameter.
!       job = 1  ---> set max(max(A)) = 1.0
!       job = 2  ---> set max of each row to be 1.0
!       job = 3  ---> scale each column so diagonal entry is 1.0
!
!     If job > 1, a vector is returned in D that gives
!     the scaling vector used, so that
!          Anew = D*A   (job = 2)  or
!          Anew = A*D   (job = 3)
!
!     In case job < 3, the right hand side vector is also scaled.
!
!     Warning: this code assumes that A is not the zero matrix!
!       Also, if the max-norm of a row is zero when job = 2, 
!       or if the diagonal element is zero when job = 3, the
!       code sets the corresponding D element to be zero and
!       does not scale that row or column.
!----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      double precision :: A(*), t
      double precision :: b(n), D(n)

      integer idamax
      external idamax

!---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
         nnz = rwptr(n+1)-1
         k = idamax(nnz, A, 1)
         t = 1.0d0/A(k)
         call dscal(nnz,t,A,1)
         call dscal(n,t,b,1)
         return
      end if 

!---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             t = abs(A(s))
             do k = s+1, rwptr(i+1)-1
               if (abs(A(k)) .gt. t) t = abs(A(k))
             end do 
             if ( t .ne. 0.0d0) then
                D(i) = 1.0d0/t
                call dscal(nnz,D(i),A(s),1)
                b(i) = D(i)*b(i)
             else
                D(i) = 1.0d0
             end if
         end do 
         return
      end if 
!---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         do i = 1, n
            D(i) = 1.0d0
            do k = rwptr(i), rwptr(i+1)-1
               if (colind(k) .eq. i .and. A(k).ne. 0.0d0) then
                  D(i) = 1.0d0/A(k)
                  go to 100
               end if
            end do
 100     continue
         end do 
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)*D(colind(i))
         end do 
         return
      end if 

      return
      end subroutine scala

!==============================================================

      subroutine unscalx(n, x, Dr)
      implicit none
      integer n, k
      double precision :: x(n), Dr(n)
      do k = 1, n
         x(k) = x(k)*Dr(k)
      end do
      return
      end subroutine unscalx

!-----------------------------------------------------------------------

      subroutine unscala(job,A, rwptr, colind, n, b, D)
      implicit none

!----------------------------------------------------------------
!     Unscale a CSR matrix to reverse effects of scala.
!----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      double precision :: A(*), t
      double precision :: b(n), D(n)

      integer idamax
      external idamax

!---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
!         nnz = rwptr(n+1)-1
!         k = idamax(nnz, A, 1)
!         t = 1.0d0/A(k)
!         call dscal(nnz,t,A,1)
!         return
      end if 

!---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             if (D(i) .ne. 0.0d0) then
                t = 1.0d0/D(i)
                call dscal(nnz,t,A(s),1)
                b(i) = t*b(i)
             end if 
         end do 
         return
      end if 
!---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)/D(colind(i))
         end do 
         return
      end if 

      return
      end subroutine unscala
!
!-----------------------------------------------------------------------
